/* 
 * orcad_convert: Lists and translates Orcad IV (and older) schematic (.sch) files
 *
 * Copyright (c) 2020 Dimitri Marinakis
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License Version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Usage:
 *	 orcad_convert [options] [.sch file] [output_file]
 *
 * Options:
 *   -h -- display program information
 *   -s -- file statistics
 *   -g -- gEDA output
 *   -k -- kicad output
 *   -m -- kicad - module connections (see MODULE PORTS, below)
 *   -n -- kicad / geda - module ports translated into wire and label nodes (see MODULE PORTS, below)
 *   -V -- processing debug information is printed
 *
 * Returns: exit code 1 if unsuccesful
 *
 * Make: gcc orcad_convert.c -o orcad_convert
 * 
 * Examples: 
 *              ./orcad_convert [-V] EXAMPLE.SCH (lists schematic contents)
 *              ./orcad_convert -k EXAMPLE.SCH example.k.sch (convert to kicad format)
 *              ./orcad_convert -g EXAMPLE.SCH example.g.sch (convert to geda format)
 * 
 *              Optional step to add library information, for kicad
 *              ./add_libraries example.k.sch
 * 
 * Issues:  
 *          ORCAD SCHEMATIC FILE STRUCTURE
 *          This work is based on reverse-engineering my own data files produced 
 *          by OrCAD L.P.'s OrCAD SDT IV/III several (...) years ago, so YMMV.
 *          Needless to say, the output should be exhaustively checked.
 *
 *          LIBRARY FILES
 *          Orcad SDT schematics depend on the presence of component libraries, in the sequence
 *          given in SDT.CFG which may or may not correspond to any specific schematic drawing.
 * 
 *          Suitable component part libraries should be available to match the components used
 *          in the original drawings. In order to allow for selection of the proper component,
 *          the following methods are used:
 * 
 *          (1) A file named 'orcad_parts.txt' is expected to be present, containing 
 *          component names, their x and y size, and correponding library name. If this file
 *          is not present, component parts may be misplaced if not in their default orientation.
 * 
 *          Information suitable for the 'orcad_parts.txt' file is produced by 'orcad_lib_convert',
 *          used to convert orcad source library files (.SRC) to kicad component libraries (.lib) or 
 *          geda symbol (.sym) files organized in directories. 
 * 
 *          Component information is in the form of individual files (e.g. TTL.SRC.txt), which
 *          should be concatenated in the library precedence order used for the schematic(s) at hand.
 * 
 *          kicad: If a schematic component name is found in the 'orcad_parts.txt' file, the
 *                 corresponding library name is prepended to the component name, e.g. TTL:74HC94
 * 
 *          kicad: Library names should match the "Nickname" in the "Manage Symbol Libraries" window, 
 *                 which is entered in the 'sym-lib-table' file present in each KiCad project directory.
 *                 e.g. (lib (name TTL)(type Legacy)(uri ...) ...)
 * 
 *          geda:  A file in the user's home directory (~/.gEDA/gafrc) should include
 *                 the following lines (replace content, as necessary):
 * 
 *                 (component-library "path_to_an_individual_component_library")
 *                 (component-library-search "path_to_a_directory_containing_symbol_directories")
 *                 (source-library "path_to_the_working_directory")
 * 
 *                 The third line is used to indicate where hierarchical drawing "source" files
 *                 - namely the files lower in the hierarchy - may be found.
 * 
 * 
 *          (2) (kicad only) Following conversion, the "add_library" script may add (or modify) the library 
 *          for each component in a kicad schematic. An example, using awk, is below:
 * 
 *              # add_library
 *              # Add component libraries to a converted
 *              # (orcad) schematic
 *              #
 *              awk '{ \
 *              gsub("L 74","L ttl:74"); \
 *              gsub("L 40","L cmos:40"); \
 *              gsub("L CAP", "L device:CAP");\
 *              gsub("L R", "L device:R");\
 *              gsub("L POT", "L device:POT");\
 *              gsub("L CRYSTAL", "L device:CRYSTAL");\
 *              print}' < $1 > $2
 * 
 *          TITLE BLOCK
 *          Title block information is provided even if the title block in the original
 *          schematic is marked 'invisible'. Date and Sheet size are always filled-in.
 * 
 *          kicad: The available title block fields are used. Document number is concatenated
 *                 with the sheet title, Address lines 1 - 4 are shown as Comment 1 to 4
 * 
 *          geda:  An embedded title block component is created at the nominal location.
 *                 If the title block interferes with other parts of the schematic, it can
 *                 be moved as a whole. FIXME: Separate attributes, if desired.
 *          
 *          SHEET - HIERARCHICAL DRAWINGS
 *          The sheet filenames in hierarchical drawings may not match the translated
 *          drawing filenames.  Edit all higher-level sheets to correct the path of the
 *          lower-level drawing filenames.
 * 
 *          geda:   Hierarchical drawings are represented by embedded components
 *                  with the "source=" attribute set.
 *          geda:   Needs to know where the sheets are located.  
 *                  The following line must be added in ~./gEDA/gafrc:
 *                  (source-library "path_to_schematic_files")
 * 
 *          COMPONENT ORIENTATION
 *          Orcad uses the "top left" point of a component, minus any pins, as the anchor.
 *          In order to properly convert the device orientation, the dimensions
 *          of each component are needed to correctly calculate any resulting translations.
 *  
 *          This information should be provided in a file called "orcad_parts.txt" 
 *          (see LIBRARY FILES, above). Absence of this file is not fatal but it will result 
 *          in misplaced components if their orientation is not the default one.
 * 
 *          Format of the "orcad_parts.txt" file contents is
 * 
 *              PART_NAME X_SIZE Y_SIZE LIBRARY
 *
 *              (sizes in mils, excluding pin length, fields separated by spaces)
 * 
 *          Part name and Reference positioning does not always match the original schematic.
 * 
 *          MULTIPART COMPONENT REFERENCES
 *          kicad: Automatically adds the part designator for multipart components (dev_num)
 *                 (e.g. 74LS04 U1C becomes U1CC)
 *                 To compensate, if the part number designator matches the corresponding 
 *                 letter of the alphabet, it is removed by the converter.
 * 
 *          POWER COMPONENTS
 *          OrCAD SDT power components (arrow, bar, circle, wave) are not library parts;
 * 
 *          kicad: A special library called 'pwr.lib' was created to compensate.
 * 
 *          geda:  Individual components in a directory called PWR were created to compensate.
 * 
 *          WIRES
 *          geda: [ IMPORTANT NOTE ]: if a wire (net) passes through pin ends it will automagically 
 *          be connected by gschem to those pins. geda does not support junctions explicitly.
 * 
 *          JUNCTIONS
 *          geda: Creates junctions automatically if a wire (net) touches a component pin.
 * 
 *          geda: Initially, the program was creating zero-length nets in place of junctions
 *          and the two-liner can be re-instated (uncommented) if it is of any help.
 * 
 *          MODULE PORTS
 *          kicad: The corresponding Global labels have one connection point. This creates a problem when 
 *                 e.g. an output port on the right of a connection is pointing to the left.
 *                 This type of port is left unconnected, unless the -m or -n options are used,
 *                 in which case the resulting schematic may look uglier but the resulting netlist 
 *                 would be closer to reality.
 * 
 *          geda:  Module ports are created as embedded components with two pins, one at each end
 *                 of the module port.
 * 
 *          geda:  Bus module ports are not supported; they are drawn but are not connected in gschem.
 * 
 *          LABELS (NET NAMES)
 *          geda: A zero-length net is created with a 'netname=' attribute in an effort to associate
 *                a label with a specific wire (net);  this may be shown in the converted schematic 
 *                with or without a junction symbol (netname shown on the corresponding wire).
 * 
 *          BUS ENTRIES
 *          Orcad SDT does not have a bus-to-bus entry; all bus entries are represented
 *          in the converted schematics by wire-to-bus entries; bus-to-bus segments should be 
 *          manually corrected.
 * 
 *          geda: Bus entries attached to the edge of the bus are left unconnected by gschem;
 *                to compensate, bus lines are lengthed by 10 mils at each edge. 
 * 
 *          TEXT
 *          Text sizes were tweaked to match the converted output representations and
 *          may not be optimal.
 * 
 *          MARKERS
 *          kicad: The No Connect marker has a corresponding directive.
 * 
 *          kicad: The Stimulus, Vector, Trace and Layout directive markers are implemented via 
 *                 components in the added 'pwr.lib' (see POWER COMPONENTS, above)
 * 
 *          kicad: The Error marker is implemented via a Text directive.
 * 
 *          geda: The PWR library directory contains the corresponding components 
 *                (see POWER COMPONENTS, above)
 * 
 * References:
 *          orcad_schematic_file_format.txt (own information for SDT III/IV, 
 *                also checked against SDT IV information from http://www.pcengines.ch)
 *          KiCad File formats
 *          gEDA/gaf File Format Document (http://wiki.geda-project.org/)
 *          gEDA/gaf Master Attribute Document
 *          gEDA_netattrib.pdf
 *          gEDA Hierarchical schematics
 *          gEDA symbols.pdf
 * 
 * History: 
 *   15-Feb-2020 dm
 *   10-Apr-2020 dm -- handling of 'older' Orcad .sch files
 *   25-Apr-2020 dm -- library information included in orcad_parts.txt file (ver. 2.2)
 *   24-May-2020 dm -- geda module ports
 */

#include "orcad_convert.h"

/****************** PROTOTYPES FROM THE TOP DOWN *******************/

int orcad_convert (char * input_file, char * output_file);
void output_statistics(char *input_f);


int process_input (int input_count);
off_t fsize(const char *filename);
int output_string (char * bstring);
void subspaces(char * string);
void subquotes(char * string);
void subcolons(char * string);
void initialize_parts_array(void);
void get_part_data(char * p);
void geda_text (int x, int y, char * p);
void geda_sheet_port(int x, int y, int seq, unsigned char *portname, char portside, char porttype); 
void process_header(void);
void process_sheet(void);
void process_component(void);
void process_wire(void);
void process_bus(void);
void process_junction(void);
void process_module_port(void);
void process_label(void);
void process_entry(void);
void process_dashed_line(void);
void process_power(void);
void process_text(void);
void process_marker(void);
void process_end_of_file(int max_iptr);

/****************** PROGRAM VERSION INFORMATION  *******************/
char *prog_version="2.9";

/****************** COMMAND LINE OPTIONS  **************************/
int opt_verbose = 0;
int opt_list = 1; /* default option */
int opt_statistics = 0;
int opt_kicad = 0;
int opt_modules = 0;
int opt_nodes = 0; /* module port conversion to wire and label nodes */
int opt_geda = 0;


/****************** GLOBAL VARIABLES *******************************/

int iptr = 0;	/* input buffer pointer, reset before every read */
int optr = 0;	/* output buffer pointer, reset after every write */
off_t infileptr = 0; /* file input pointer, updated with last input - print with %lld */
off_t infilesize = 0; /* input file size */
unsigned char input_buffer[INRECSIZE];
unsigned char output_buffer[OUTRECSIZE];

int processing = 0; /* processing flag may be modified by individual processing routines */
int header_processed = 0; /* Header will be processed first */
int total_sheets = 0;
int total_components = 0;
int total_wires = 0;
int total_busses = 0;
int total_junctions = 0;
int total_module_ports = 0;
int total_labels = 0;
int total_entries = 0;
int total_dashed_lines = 0;
int total_power_entries = 0;
int total_text_entries = 0;
int total_markers = 0;
int total_items = 0;
uint32_t components;  /* components + 0x20 or 0x1b for older version = file offset of component list */

unsigned char parts[MAXCOMPONENTS][MAXSTRING];
int part_x_size[MAXCOMPONENTS];
int part_y_size[MAXCOMPONENTS];
unsigned char part_libraries[MAXCOMPONENTS][MAXSTRING];

int x_size, y_size; /* component size, if found, else zero */
unsigned char part_library[MAXSTRING];

int version; /* this is 2 for SDT IV and 1 for older version */

int geda_x = 0; /* geda X is sheet extent */
int geda_y = 0; /* geda Y axis starts at bottom of sheet and goes up */

/****************** HANDLE ARGUMENTS AND SYNTAX *******************/

void usage_info(void)
{
	printf("\norcad_convert: Schematic (.SCH) data dumper and converter ver. %s\n", prog_version);
	printf("\norcad_convert: Copyright (c) 2020 Dimitri Marinakis");
	printf("\norcad_convert: This program is free software; you are encouraged to redistribute it under");
	printf("\norcad_convert: the terms of the GNU General Public License (version 2).\n");
	printf("\norcad_convert: This program comes with ABSOLUTELY NO WARRANTY. See the GNU General Public");
	printf("\norcad_convert: License in the file named `COPYING' for more details.\n");
	printf("\norcad_convert: Syntax: orcad_convert [options] [.sch file] [output_file]\n");
	printf("\norcad_convert:         (if .sch_file is not given, program reads from standard input)\n");
	printf("\norcad_convert: -h -- display program information");
    printf("\norcad_convert: -g -- geda output");
	printf("\norcad_convert: -k -- kicad output");
	printf("\norcad_convert: -m -- kicad 'shorted' module ports");
    printf("\norcad_convert: -n -- kicad or geda - module ports translated into wire and label nets");
	printf("\norcad_convert: -s -- statistics");
	printf("\norcad_convert: -V -- additional processing information");
	printf("\n\n");
}

int main(int argc, char * argv[])
{
	unsigned char ucc;	/* test variable */
	int idx, i;

	while(argc > 1 && argv[1][0] == '-') {
		switch(argv[1][1]) {
			case 'V':
				opt_verbose = 1; /* careful - will produce additional output */
				break;
			case 's':
				opt_statistics = 1;
				break;
            case 'g':
                opt_geda = 1;
                break;
            case 'k':
                opt_kicad = 1;
                break;
            case 'm':
                opt_modules = 1;
                break;
            case 'n':
                opt_nodes = 1;
                break;
			case '?':
			case 'h':
//				opt_xxx = 1;
//				break;
// goes through if break is not uncommented
			default:
				usage_info();
				exit(0);
		}
	argc-- ;
	argv++ ;
	}
	
	/* Resolve incompatible options
     * kicad takes precedence over geda
     * NOTE: -V (verbose output) is left untouched - and will produce
     *       output not readily understood by the drawing applications
     */
    if (opt_kicad) {
        opt_geda = 0;
    }
    if (opt_nodes) {
        opt_modules = 0;
    }
    if (opt_kicad || opt_geda) {
        opt_list = 0;
    }

	switch(argc) {
		case 1:
			return orcad_convert("", "");
			break;
		case 2:
			return orcad_convert(argv[1], "");
			break;
		default:
			return orcad_convert(argv[1], argv[2]);
 	}
}


/****************** FILE READ-WRITE LOOP **************************/

int orcad_convert(char *input_file, char *output_file)
{
	int i;      /* counter */
	int j;      /* counter */
	int infile; /* input file descriptor */
	int outfile;/* output file descriptor */

	int icnt;   /* input file bytes read in input buffer */
	int ocnt;   /* output file bytes written */
	int bytes_to_process;	/* bytes read minus bytes already processed */

	int wehaveinput;    /* flag for while */
	int process_return;    /* process return code */

	char new_file[MAXFILELEN];
	struct stat filestat;  /* used to obtain the file size */
	
	/* Open input and output files
	 */
// 	infile = open(input_file, O_RDONLY);
	if (strlen(input_file) == 0) {
		infile = STDIN_FILENO;
	} else {
		infile = open(input_file, O_RDONLY);
        infilesize = fsize(input_file);
	}
//printf("\norcad_convert: #1\n");
	if (infile < 0 || infilesize < 0) {
		perror("\norcad_convert: input file open");
		return(1);
	} else {
		if (strlen(output_file) == 0) {
			outfile = STDOUT_FILENO;
		} else {
			if (strlen(output_file) < MAXFILELEN) {
				strcpy(new_file, output_file);
			} else {
				printf("\norcad_convert: output filename too long - exiting\n");
				return(1);
			}
			outfile = open(new_file, O_WRONLY | O_CREAT | O_TRUNC);
		}
		if (outfile < 0) {
			perror("\norcad_convert: output file create");
			close(infile);
			return(1);
		}
	}

	/* Initialize processing variables
	 */
	total_items = 0;
    initialize_parts_array();

	/*  Read, process and write file blocks,
	 *  Change file mode (equivalent to chmod 644 output_file),
	 *  and return.
     * 
     *  (NOTE: this version uses a one-shot read)
     * 
	 */
	wehaveinput = 1;
	while (wehaveinput) {
//printf("\norcad_convert: #3\n");
		icnt = read(infile, input_buffer, sizeof(input_buffer));
		if (icnt == 0) wehaveinput = 0;
//printf("\norcad_convert: #4 wehaveinput: %d\n", wehaveinput);
        
        /* Processing loop */
		iptr = 0;
		while ((icnt > 0) && (iptr < icnt)) {
			bytes_to_process = icnt - iptr;
			process_return = process_input(bytes_to_process);          
//printf("\norcad_convert: optr: %d\n", optr);
			/* Write processed data and reset output buffer pointer */
			if (optr >= 0) { /* processing could return with zero output bytes */
				ocnt = write(outfile, output_buffer, optr);
				optr = 0;
				if (ocnt < 0) {
					perror("\norcad_convert: output file write");
					wehaveinput = 0;
				}
			} else if (process_return != -2) {
				if (opt_verbose) printf("\norcad_convert: no more bytes to write");
				wehaveinput = 0;    /* signal no more input */
			}
		}
	}

	/* Output statistics, if requested,
	 * close input and output files,
	 * make output file readable.
	 */
	close(infile);
	if (opt_statistics || opt_verbose) output_statistics(input_file);
	if (opt_verbose) printf("\norcad_convert: processing complete\n");
	if (outfile == STDOUT_FILENO) printf("\n");

	if (close(outfile)) {
		perror("\norcad_convert: output file close");
		return(1);
	}
	if (outfile != STDOUT_FILENO) {
		if (chmod(new_file, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) {
			perror("\norcad_convert: output file chmod");
			return(1);
		}
	}
	return(0);
}

/****************** PROCESSING *************************************/

/* process_input:
 * Processes <input_count> bytes in <input_buffer>
 * Returns: bytes written, -1 for EOF
 * Changes: iptr, infileptr
 */
int process_input (int input_count)
{
	unsigned char inchar;
	unsigned int outcode;
	unsigned char current;
	int iptr_max;       /* holds the calculated maximum input pointer value */
	int return_code;    /* bytes written; error if negative */

	return_code = 0;

	processing = 1;
	iptr_max = iptr + input_count;

	if (opt_verbose) printf("\n\norcad_convert: process_input - %d bytes, iptr = %4.4x, iptr_max = %4.4x\n", input_count, iptr, iptr_max);
    
    while (processing) {
		if ((iptr < INRECSIZE) && (iptr < iptr_max)) {
			inchar = input_buffer[iptr++];
            if (opt_verbose) printf("\n\norcad_convert: process_input before %d bytes, TAG = %2x, iptr = %4.4x, iptr_max = %4.4x, infileptr = %4.4x\n", input_count, inchar, iptr, iptr_max, infileptr);
//            printf("*");
            inchar = inchar & 0xF;
            if (header_processed == 0) { process_header();
            } else if (inchar == SHEET) { process_sheet();
            } else if (inchar == COMPONENT) { process_component();
            } else if (inchar == WIRE) { process_wire();
            } else if (inchar == BUS) { process_bus();
            } else if (inchar == JUNCTION) { process_junction();
            } else if (inchar == MODULE_PORT) { process_module_port();
            } else if (inchar == LABEL) { process_label();
            } else if (inchar == ENTRY) { process_entry();
            } else if (inchar == DASHED_LINE) { process_dashed_line();
            } else if (inchar == POWER) { process_power();
            } else if (inchar == TEXT) { process_text();
            } else if (inchar == MARKER) { process_marker();
            } else if (inchar == END_OF_FILE) { process_end_of_file(iptr_max);
            } else { printf("\n\norcad_convert: process_input unknown tag - %d bytes, TAG = %2x, iptr = %4.4x, iptr_max = %4.4x, infileptr = %4.4x\n", input_count, inchar, iptr, iptr_max, infileptr);
            }
		} else {
			/* Finished processing all input */
            if (opt_list) printf("\n");
			processing = 0;
		}
	} /* end while processing*/

	if (opt_verbose) printf("\norcad_convert: iptr - %4.4x, optr - %4.4x ", iptr, optr);

	return return_code;
}/* iptr should be zero to begin with */



/****************** LIBRARY FUNCTIONS ******************************/
off_t fsize(const char *filename) {
    struct stat st;

    if (stat(filename, &st) == 0)
        return st.st_size;

    fprintf(stderr, "orcad_convert: Cannot determine size of %s: %s\n",
            filename, strerror(errno));

    return -1;
}

int output_string (char * bstring)
{
	/* Writes strlen(bstring) bytes to output_buffer
	 * Modifies optr
	 */
	int i;
	int len;
    int return_code;
    
    return_code = 0;
	len = strlen(bstring);
	if (len && (optr < OUTRECSIZE)) {
		for (i=0; i<len; i++) output_buffer[optr++] = bstring[i];
	} else {
        return_code = -1;
    }
    return return_code;
}

void subspaces(char * string)
{
   /* Substitutes spaces, hyphens ('-') and slashes ('/') with '_' for component names
    * kicad: spaces are not allowed in component names
    * geda: hyphens and slashes are not allowed in component names
    */
    for (int i=0; i<strlen(string); i++) {
        if(string[i] == 0x20 || string[i] == 0x2d || string[i] == 0x2f) {
            string[i] = '_';
        }
    }
}

void subquotes(char * string)
{
   /* Substitutes quotes with '/' (for kicad pin names)
    */
    for (int i=0; i<strlen(string); i++) {
        if(string[i] == 0x22) {
            string[i] = '/';
        }
    }
}

void subcolons(char * string)
{
   /* Substitutes colons with '_' (for gEDA power components)
    */
    for (int i=0; i<strlen(string); i++) {
        if(string[i] == 0x3A) {
            string[i] = '_';
        }
    }
}

void initialize_parts_array(void)
{
    /* Reads a parts file with 
     *    part name, x size, y size, part library
     * on each line, separated by spaces.
     * 
     * If a parts file is not found, the corresponding information will be missing
     * which could result in misplaced parts on the drawing.
     */
    FILE *comp;
#define MAXTOKENS 6
    unsigned char input_line[MAXSTRING];
    unsigned char tokens[MAXTOKENS][MAXSTRING];
    int tokenidx;
    char *token;
    int i;
    for (int i=0; i<MAXCOMPONENTS; i++) {
        parts[i][0] = '\0';
        part_x_size[i] = 0;
        part_y_size[i] = 0;
        part_libraries[i][0] = '\0';
    }
    comp = fopen("./orcad_parts.txt", "r"); /* hard-coded input file */
    if (comp == NULL) {
		perror("\norcad_convert: cannot open orcad_parts.txt");
    } else {
    i = 0;
    while (fgets(input_line, MAXSTRING-1, comp) != NULL) {
        /* break line into tokens - expected are: 
         * component name, x size, y size, part library
         */
        tokenidx = 0;
        token = strtok(input_line, " "); 
        if (token != NULL) strcpy(tokens[tokenidx++], token);
        while (token != NULL && tokenidx <= MAXTOKENS) { 
            token = strtok(NULL, " ");
            if (token != NULL) strcpy(tokens[tokenidx++], token);
        }
        strncpy(parts[i], tokens[0], MAXSTRING-1);
        part_x_size[i] = atoi(tokens[1]);
        part_y_size[i] = atoi(tokens[2]);
        strncpy(part_libraries[i], tokens[3], MAXSTRING-1);
        i++;
    }
        fclose(comp);
    }
}

void get_part_data(char * p)
{
    /* Search for part in constant array  */
    int i=0;
    x_size = 0;
    y_size = 0;
    part_library[0] = '\0';
    while (strlen(parts[i]) != 0) {
        if (strcmp(parts[i], p) == 0) {
            x_size = part_x_size[i];
            y_size = part_y_size[i];
            strncpy(part_library, part_libraries[i], MAXSTRING-1);
            part_library[strlen(part_library)-1] = '\0'; /* replace EOL with null */
            break;
        }
        i++;
    }
}

void geda_text (int x, int y, char * p)
{
    unsigned char outstr[MAXSTRING];
    sprintf(outstr, "\nT %d %d 6 %d 1 0 0 0 1\n%s", \
    x, y, 6, p);
        output_string(outstr);
}

void geda_sheet_port(int x, int y, int seq, unsigned char *portname, char portside, char porttype)
{
    /* outputs (zero length) pin information as part of an
     * embedded component
     */
    unsigned char tstr[MAXSTRING];
    int alignment = 0;
    
    /* Draw a pin symbol */
    sprintf(tstr,"\nH 3 10 0 0 -1 -1 0 -1 -1 -1 -1 -1 3");
    output_string(tstr);
    sprintf(tstr,"\nM %d %d", x, y);
    output_string(tstr);
    if (porttype == 'O') {
        /* Output port */
        if (portside == 'R') {
        sprintf(tstr,"\nl %d,%d %d,%d %d,%d %d,%d\nz", \
            -30,30, -30,0, 0,-60, 30,0);
        } else {
            sprintf(tstr,"\nl %d,%d %d,%d %d,%d %d,%d\nz", \
            30,30, 30,0, 0,-60, -30,0);
        }
    } else {
        /* Input port */
        if (portside == 'R') {
            sprintf(tstr,"\nl %d,%d %d,%d %d,%d %d,%d %d,%d\nz", \
                0,30, -30,0, -30,-30, 30,-30, 30,0);
        } else {
            sprintf(tstr,"\nl %d,%d %d,%d %d,%d %d,%d %d,%d\nz", \
            0,30, 30,0, 30,-30, -30,-30, -30,0);
        }
    }
    output_string(tstr);
    
    sprintf(tstr, "\nP %d %d %d %d 1 0 0\n{", x, y, x, y);
    output_string(tstr);
    
    /* labels are placed with a slight offset inside the sheet component
     */
    if (portside == 'L') {
        x = x + 80;
        alignment = 1; 
    } else { 
        x = x - 80;
        alignment = 7;
    }
    sprintf(tstr, "\nT %d %d 5 6 0 1 0 %d 1", x, y, alignment);
    output_string(tstr);
    sprintf(tstr, "\npinseq=%d", seq);
    output_string(tstr);
    sprintf(tstr, "\nT %d %d 5 6 0 1 0 %d 1", x, y, alignment);
    output_string(tstr);
    sprintf(tstr, "\npinnumber=%d", seq);
    output_string(tstr);
    sprintf(tstr, "\nT %d %d 5 6 0 1 0 %d 1", x, y, alignment);
    output_string(tstr);
    sprintf(tstr, "\npintype=%s", (porttype == 'O')?"out":"in");
    output_string(tstr);
    sprintf(tstr, "\nT %d %d 5 6 1 1 0 %d 1", x, y, alignment);
    output_string(tstr);
    sprintf(tstr, "\npinlabel=%s", portname);
    output_string(tstr);
    sprintf(tstr, "\n}");
    output_string(tstr);
}


/****************** PROCESSING FUNCTIONS ******************************/

/*
 * Processing functions are entered with iptr pointing to the byte count.
 * the tag byte and the byte count should be added to infileptr (+3 bytes)
 * while only the byte count (2 bytes) should be added to iptr
 */

/* process_header
 * Header length is 0x172 bytes (or 0x16d bytes for older version)
 * and it is the first item read.
 * Header includes title block (Tag = 0)
 * 
 * kicad variable names are used for Revision, Organization, Address
 * in order to avoid extensive duplication. 
 */
void process_header(void)
{
    unsigned char stemp[MAXSTRING];
    unsigned char title[MAXSTRING];
    unsigned char docnum[MAXSTRING/2]; /* this is appended to title, if present */
    
    int temp;
    int sheet, number_of_sheets;
    int tempaddr;
    int geda_title_offset;
    
    if ( (iptr == 1) && ((iptr + HEADER_RECORD_LENGTH) < INRECSIZE) ) {
        
        memcpy(&stemp, &input_buffer[0], 0x10); /* Schematic FILE */
        stemp[0x10] = '\0';

        /* Distinguish between older and newer Orcad SDT versions
         * (newer version has a 5-byte offset with respect to the old one)
         */
        version = 0;
        if (strncmp((char *) stemp,"Schematic FILE", 14) == 0) {
            version = 2;
        } else if (strncmp((char *) stemp,"SCHEMATIC FILE", 14) == 0) {
            version = 1;
        }
        if (version == 0) {
            printf("\norcad_convert: missing 'Schematic FILE' magic!\n\n");
            processing = 0; /* end processing */
        } else {
            if (opt_kicad) sprintf(stemp,"EESchema Schematic File Version 4");
            else if (opt_geda) sprintf(stemp,"v 20130925 2");
            output_string(stemp);
        }
        if (processing != 0) {
            tempaddr = (version == 2)?0x16:0x11;
            memcpy(&components, &input_buffer[tempaddr], 4);
            sprintf(stemp, "\nComponent list offset (hex, dec): %4.4x %u\n", components, components);
            /* 
             * NOTE: The LIBS and EELAYER directives are not currently used by kicad
             * NOTE: 'pwr.lib' is an added component library containing Orcad SDT - like symbols and it should be used with the rendered schematic
             */
//            if (opt_kicad) sprintf(stemp, "\nLIBS: pwr");
//            output_string(stemp);
            components = (version == 2)?components + 0x20: components + 0x1b; /* Point to first component part name */
            
            if (opt_kicad) {
                sprintf(stemp, "\nEELAYER 29 0\nEELAYER END");
                output_string(stemp);
            }
            tempaddr = (version == 2)?0x1a:0x15;
            temp = 0; memcpy(&temp, &input_buffer[tempaddr], 2);
            sprintf(stemp, "\nCurrent x cursor position (hex, dec): %4.4x %d\n", temp, temp);
            if (opt_list) output_string(stemp);
            
            tempaddr = (version == 2)?0x1c:0x17;
            temp = 0;  memcpy(&temp, &input_buffer[tempaddr], 2);
            sprintf(stemp, "\nCurrent y cursor position (hex, dec): %4.4x %d\n", temp, temp);
            if (opt_list) output_string(stemp);
            
            tempaddr = (version == 2)?0x1e:0x19;
            temp = 0;  memcpy(&temp, &input_buffer[tempaddr], 1);
            sprintf(stemp, "\nCurrent zoom level: %d\n", temp);
            if (opt_list) output_string(stemp);
            
            /* Title block */
            tempaddr = (version == 2)?0x23:0x1e;
            temp = 0;  memcpy(&temp, &input_buffer[tempaddr], 2);
            sprintf(stemp, "\nSheet number (0 = invisible): %d\n", temp);
            if (opt_list) output_string(stemp); 
            sheet = temp;
            
            tempaddr = (version == 2)?0x25:0x20;
            temp = 0; memcpy(&temp, &input_buffer[tempaddr], 2);
            sprintf(stemp, "\nNumber of sheets (0 = invisible) %d\n", temp);
            if (opt_list) output_string(stemp); 
            number_of_sheets = temp;
            
            tempaddr = (version == 2)?0x27:0x22;
            temp = 0;  memcpy(&temp, &input_buffer[tempaddr], 1);
            if ((temp & 0x8) && opt_list) sprintf(stemp, "\nInvisible title block\n");
            if (opt_geda) {
                /* Create an embedded title component for geda - starts here */
                sprintf(stemp, "\nC %d %d 1 0 0 EMBEDDEDTITLE_BLOCK_GEDA-1.sym\n[", geda_x, geda_y);
                output_string(stemp);
            }
            if (opt_list) {
                sprintf(stemp, "\nSheet size ", temp);
                output_string(stemp);
            }
            if (opt_kicad) {
                sprintf(stemp, "\n$Descr ");
                output_string(stemp);
            }
            switch (temp &= 0x0f) {
                case 0:
                    sprintf(stemp, "A 11000 8500");
                    geda_x = 11000;
                    geda_y = 8500;
                    break;
                case 1:
                    sprintf(stemp, "B 17000 11000");
                    geda_x = 17000;
                    geda_y = 11000;
                    break;
                case 2:
                    sprintf(stemp, "C 22000 17000");
                    geda_x = 22000;
                    geda_y = 17000;
                    break;                    
                case 3:
                    sprintf(stemp, "D 34000 22000");
                    geda_x = 34000;
                    geda_y = 22000;
                    break;
                case 4:
                    sprintf(stemp, "E 44000 34000");
                    geda_x = 44000;
                    geda_y = 34000;
                    break;
                default:
                    sprintf(stemp, "User 48000 36000"); /* arbitrary dimensions - larger than ISO E */
                    geda_x = 48000;
                    geda_y = 36000;
            }
            if (opt_list || opt_kicad) output_string(stemp); /* sheet size */
            if (opt_geda) {
                geda_text(geda_x - 5070, 1800, "Size"); /* sheet size */
                stemp[1] = '\0'; /* only need to print the sheet size letter */
                geda_text(geda_x - 5030, 1650, stemp);
            }
            
            if (opt_kicad) {
                sprintf(stemp, "\nencoding utf-8"); 
                output_string(stemp);
                sprintf(stemp, "\nSheet %d %d", sheet, number_of_sheets); 
                output_string(stemp);
            } else if (opt_geda) {
                sprintf(stemp, "Sheet %d of %d", sheet, number_of_sheets); 
                geda_text(geda_x - 3000, 1510, stemp); /* sheet size */
            }

            if (opt_list || opt_kicad) {
                sprintf(stemp, "\nDate \"");
                output_string(stemp);
            }
            tempaddr = (version == 2)?0x28:0x23;
            temp = 0;  memcpy(&temp, &input_buffer[tempaddr], 1);
            memcpy(&stemp, &input_buffer[tempaddr+1], temp); 
            stemp[temp] = '\"';
            stemp[temp+1] = '\0';
            if (opt_list || opt_kicad) output_string(stemp);
            if (opt_geda) {
                geda_text(geda_x - 5080, 1510, "Date:");
                stemp[temp] = '\0'; /* don't need the closing quotes */
                geda_text(geda_x - 4700, 1510, stemp); /* Date */
            }
            
            sprintf(stemp, "\nDocument Number: \""); /* does not exist in kicad */
            if (opt_list) output_string(stemp);
            tempaddr = (version == 2)?0x3b:0x36;
            temp = 0;  memcpy(&temp, &input_buffer[tempaddr], 1);
            memcpy(&stemp, &input_buffer[tempaddr+1], temp); 
            stemp[temp] = '\"';
            stemp[temp+1] = '\0';
            if (opt_list) output_string(stemp);
            strncpy(docnum, stemp, strlen(stemp)-1);
            docnum[strlen(stemp)-1] = '\0';
            
            if (opt_list || opt_kicad) {
                sprintf(stemp, "\nRev \"");
                output_string(stemp);
            }
            tempaddr = (version == 2)?0x60:0x5b;
            temp = 0;  memcpy(&temp, &input_buffer[tempaddr], 1);
            memcpy(&stemp, &input_buffer[tempaddr+1], temp); 
            stemp[temp] = '\"';
            stemp[temp+1] = '\0';
            if (opt_list || opt_kicad) output_string(stemp);
            if (opt_geda) {
                geda_text(geda_x - 1780, 1800, "REV");
                stemp[temp] = '\0'; /* don't need the closing quotes */
                geda_text(geda_x - 1750, 1650, stemp); /* Revision */
            }
            
            if (opt_list) {
                sprintf(stemp, "\nTitle: \"");
                output_string(stemp);
            }
            tempaddr = (version == 2)?0x64:0x5f;
            temp = 0;  memcpy(&temp, &input_buffer[tempaddr], 1);
            memcpy(&stemp, &input_buffer[tempaddr+1], temp);
            stemp[temp] ='\"';
            stemp[temp+1] = '\0';
            if (opt_list) output_string(stemp);
            strcpy(title, stemp);
            strncpy(title, stemp, strlen(stemp)-1);
            title[strlen(stemp)-1] = '\0';
            
            if (opt_kicad) {
                if (strlen(docnum)) {
                    sprintf(stemp, "\nTitle \"%s | %s\"", title, docnum);
                } else {
                    sprintf(stemp, "\nTitle \"%s\"", title);
                }
                output_string(stemp);
            } else if (opt_geda) {
                geda_text(geda_x - 5000, 2000, title); /* Title */
                geda_text(geda_x - 4730, 1800, "Document Number");
                geda_text(geda_x - 4730, 1650, docnum);
            }

            if (opt_list || opt_kicad) {
                sprintf(stemp, "\nComp \""); /* Organization Name */
                output_string(stemp);
                tempaddr = (version == 2)?0x91:0x8c;
                temp = 0;  memcpy(&temp, &input_buffer[tempaddr], 1);
                memcpy(&stemp, &input_buffer[tempaddr+1], temp); 
                stemp[temp] = '\"';
                stemp[temp+1] = '\0';
                output_string(stemp);
                
                sprintf(stemp, "\nComment1 \""); /* Address Line 1 */
                output_string(stemp);
                tempaddr = (version == 2)?0xbe:0xb9;
                temp = 0;  memcpy(&temp, &input_buffer[tempaddr], 1);
                memcpy(&stemp, &input_buffer[tempaddr+1], temp);
                stemp[temp] = '\"';
                stemp[temp+1] = '\0';
                output_string(stemp);

                sprintf(stemp, "\nComment2 \""); /* Address Line 2 */
                output_string(stemp);
                tempaddr = (version == 2)?0xeb:0xe6;
                temp = 0;  memcpy(&temp, &input_buffer[tempaddr], 1);
                memcpy(&stemp, &input_buffer[tempaddr+1], temp);
                stemp[temp] = '\"';
                stemp[temp+1] = '\0';
                output_string(stemp);
                
                sprintf(stemp, "\nComment3 \""); /* Address Line 3 */
                output_string(stemp);
                tempaddr = (version == 2)?0x118:0x113;
                temp = 0;  memcpy(&temp, &input_buffer[tempaddr], 1);
                memcpy(&stemp, &input_buffer[tempaddr+1], temp); 
                stemp[temp] = '\"';
                stemp[temp+1] = '\0';
                output_string(stemp);

                sprintf(stemp, "\nComment4 \""); /* Address Line 4 */
                output_string(stemp);
                tempaddr = (version == 2)?0x145:0x140;
                temp = 0;  memcpy(&temp, &input_buffer[tempaddr], 1);
                memcpy(&stemp, &input_buffer[tempaddr+1], temp);
                stemp[temp] = '\"';
                stemp[temp+1] = '\0';;
                output_string(stemp);
            } else if (opt_geda) {            
                tempaddr = (version == 2)?0x91:0x8c;
                temp = 0;  memcpy(&temp, &input_buffer[tempaddr], 1);
                memcpy(&stemp, &input_buffer[tempaddr+1], temp); 
                stemp[temp] = '\0';
                geda_text(geda_x - 5000, 2750, stemp); /* Company */

                tempaddr = (version == 2)?0xbe:0xb9;
                temp = 0;  memcpy(&temp, &input_buffer[tempaddr], 1);
                memcpy(&stemp, &input_buffer[tempaddr+1], temp);
                stemp[temp] = '\0';
                geda_text(geda_x - 5000, 2600, stemp); /* Address Line 1 */

                tempaddr = (version == 2)?0xeb:0xe6;
                temp = 0;  memcpy(&temp, &input_buffer[tempaddr], 1);
                memcpy(&stemp, &input_buffer[tempaddr+1], temp);
                stemp[temp] = '\0';
                geda_text(geda_x - 5000, 2500, stemp); /* Address Line 2 */
                
                tempaddr = (version == 2)?0x118:0x113;
                temp = 0;  memcpy(&temp, &input_buffer[tempaddr], 1);
                memcpy(&stemp, &input_buffer[tempaddr+1], temp); 
                stemp[temp] = '\0';
                geda_text(geda_x - 5000, 2400, stemp); /* Address Line 3 */

                tempaddr = (version == 2)?0x145:0x140;
                temp = 0;  memcpy(&temp, &input_buffer[tempaddr], 1);
                memcpy(&stemp, &input_buffer[tempaddr+1], temp);
                stemp[temp] = '\0';
                geda_text(geda_x - 5000, 2300, stemp); /* Address Line 4 */
                
                /* Draw title block, from the bottom up 
                 */
                /* Horizontal lines */
                sprintf(stemp,"\nL %d %d %d %d 7 0 0 0 -1 -1", \
                    geda_x - 5100, 1500, geda_x - 1500, 1500);
                output_string(stemp);
                sprintf(stemp,"\nL %d %d %d %d 7 0 0 0 -1 -1", \
                    geda_x - 5100, 1610, geda_x - 1500, 1600);
                output_string(stemp);
                sprintf(stemp,"\nL %d %d %d %d 7 0 0 0 -1 -1", \
                    geda_x - 5100, 1900, geda_x - 1500, 1900);
                output_string(stemp);
                sprintf(stemp,"\nL %d %d %d %d 7 0 0 0 -1 -1", \
                    geda_x - 5100, 2200, geda_x - 1500, 2200);
                output_string(stemp);
                sprintf(stemp,"\nL %d %d %d %d 7 0 0 0 -1 -1", \
                    geda_x - 5100, 2900, geda_x - 1500, 2900);
                output_string(stemp);
                /* Vertical lines */
                sprintf(stemp,"\nL %d %d %d %d 7 0 0 0 -1 -1", \
                    geda_x - 5100, 1500, geda_x - 5100, 2900);
                output_string(stemp);
                sprintf(stemp,"\nL %d %d %d %d 7 0 0 0 -1 -1", \
                    geda_x - 1500, 1500, geda_x - 1500, 2900);
                output_string(stemp);
                sprintf(stemp,"\nL %d %d %d %d 7 0 0 0 -1 -1", \
                    geda_x - 4750, 1610, geda_x - 4750, 1900);
                output_string(stemp);
                sprintf(stemp,"\nL %d %d %d %d 7 0 0 0 -1 -1", \
                    geda_x - 3150, 1500, geda_x - 3150, 1610);
                output_string(stemp);
                sprintf(stemp,"\nL %d %d %d %d 7 0 0 0 -1 -1", \
                    geda_x - 1800, 1600, geda_x - 1800, 1900);
                output_string(stemp);
                /* Mark end of embedded geda component */
                sprintf(stemp, "\n]");
                output_string(stemp);
                
            }
            
            if (opt_kicad) {
                sprintf(stemp, "\n$EndDescr"); 
                output_string(stemp);
            }
            if (opt_list) { 
                sprintf(stemp, "\n");
                output_string(stemp);
            }
            
            infileptr = (version == 2)? HEADER_RECORD_LENGTH: HEADER_RECORD_LENGTH-5;
            iptr = (version == 2)? HEADER_RECORD_LENGTH: HEADER_RECORD_LENGTH-5;
            
            header_processed = 1;
        }
    } else {
        processing = 0;
    }
}

/* process_sheet (Tag 01)
 * variable length
 */
void process_sheet(void)
{
    int byte_count = 0;

    unsigned char stemp[MAXSTRING];
    unsigned char sheet_filename[MAXSTRING];
    unsigned char sheet_name[MAXSTRING];
    unsigned char port_name[MAXSTRING];
    int temp;
    int ptr;
    int nets, i;
    int pos_x_top, pos_y_top, width, height;
    int port_side, port_offset, port_type, port_x, port_y;
    int uniq_1, uniq_2;
    
    ptr = iptr + 2; /* Initialize local pointer - skip byte count */
    
    memcpy(&byte_count, &input_buffer[iptr], 2);
    iptr = iptr + byte_count + 2;
    infileptr = infileptr + byte_count + 3;
    total_sheets++;
    
    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    if (opt_list) {
        sprintf(stemp, "\nSheet position top left x: %d", temp);
        output_string(stemp);
    }
    pos_x_top = temp * SCALE + OFF_X;
    ptr += 2;

    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    if (opt_list) {
        sprintf(stemp, "\nSheet position top left y: %d", temp);
        output_string(stemp);
    }
    pos_y_top = temp * SCALE + OFF_Y;
    if (opt_geda) pos_y_top = geda_y - pos_y_top;
    ptr += 2;

    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    if (opt_list) {
        sprintf(stemp, "\nSheet width: %d", temp);
        output_string(stemp);
    }
    width = temp * SCALE;
    ptr += 2;
    
    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    if (opt_list) {
        sprintf(stemp, "\nSheet height: %d", temp);
        output_string(stemp);
    }
    height = temp * SCALE;
    ptr += 2;
    
    if (opt_kicad) {
        sprintf(stemp, "\n$Sheet\nS %d %d %d %d", pos_x_top, pos_y_top, width, height);
        output_string(stemp);
    }
    if (version == 2) {
        temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
        if (opt_list) {
            sprintf(stemp, "\nSheet Unique ID 1 (hex): %x", temp);
            output_string(stemp);
        }
        uniq_1 = temp;
        ptr += 2;
        
        temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
        if (opt_list) {
            sprintf(stemp, "\nSheet Unique ID 2 (hex): %x", temp);
            output_string(stemp);
        }
        uniq_2 = temp;
        ptr += 2;
    } else {
        uniq_2 = 0;
        uniq_1 = 0;
    }
    
    if (opt_kicad) {
        sprintf(stemp, "\nU %4.4X%4.4X", uniq_2, uniq_1);
        output_string(stemp);
    } else if (opt_geda) {
        /* embedded sheet component start
         */
        sprintf(stemp, "\nC %d %d 1 0 0 EMBEDDED%4.4X%4.4X-1.sym\n[", pos_x_top, pos_y_top - height, uniq_2, uniq_1);
        output_string(stemp);
        sprintf(stemp, "\nB %d %d %d %d 3 5 0 0 75 50 0 -1 -1 -1 -1 -1", pos_x_top, pos_y_top - height, width, height);
        output_string(stemp);
    }
    
    temp = 0; memcpy(&temp, &input_buffer[ptr], 1);
    if (opt_list) {
        sprintf(stemp, "\nSheet nets: %d", temp);
        output_string(stemp);
    }
    ptr += 1;
    nets = temp;
    if (opt_list) {
        sprintf(stemp, "\nSheet filename: ");
        output_string(stemp);
    }
    temp = 0; memcpy(&temp, &input_buffer[ptr], 1);
    ptr++;
    memcpy(&stemp, &input_buffer[ptr], temp);
    stemp[temp] = '\0';
    if (opt_list) output_string(stemp);
    ptr = ptr + temp;
    strcpy(sheet_filename, stemp);

    if (opt_list) {
        sprintf(stemp, "\nSheet name: ");
        output_string(stemp);
    }
    temp = 0; memcpy(&temp, &input_buffer[ptr], 1);
    ptr++;
    memcpy(&stemp, &input_buffer[ptr], temp);
    stemp[temp] = '\0';
    if (opt_list) output_string(stemp);
    ptr = ptr + temp;
    strcpy(sheet_name, stemp);
    
    if (opt_kicad) {
        sprintf(stemp, "\nF0 \"%s\" 60", sheet_name); /* sheet name - text size in mils */
        output_string(stemp);
        sprintf(stemp, "\nF1 \"%s\" 60", sheet_filename); /* text size in mils */
        output_string(stemp);
    }

    /* Handle sheet nets */

    for (i=1; i<=nets; i++) {
    //    printf("\n%d", i);
            temp = 0; memcpy(&temp, &input_buffer[ptr], 1);
            if (opt_list) {
                sprintf(stemp, "\nSheet net %d unknown (hex): %x", i, temp);
                output_string(stemp);
            }
            ptr++;

            temp = 0; memcpy(&temp, &input_buffer[ptr], 1);
            if (temp == 0x00) {
                sprintf(stemp, "\nSheet net %d at: left", i);
                port_side = 'L';
            } else if (temp == 0xff) {
                sprintf(stemp, "\nSheet net %d at: right", i);
                port_side = 'R';
            }
            if (opt_list) output_string(stemp);
            ptr++;
            
            temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
            if (opt_list) {
                sprintf(stemp, "\nSheet net %d offset from top: %d", i, temp);
                output_string(stemp);
            }
            port_offset = (temp * SCALE) - 40; /* HACK: to align, 40 mils had to be subtracted */
            ptr += 2;
            
            temp = 0; memcpy(&temp, &input_buffer[ptr], 1);
            if (opt_list) {
                sprintf(stemp, "\nSheet net %d type:", i, temp);
                output_string(stemp);
            }
            
            switch (temp & 0x3) {
                case 0x00:
                    sprintf(stemp, "unspecified");
                    port_type = 'U';
                    break;
                case 0x01:
                    sprintf(stemp, "output");
                    port_type = 'O';
                    break;
                case 0x02:
                    sprintf(stemp, "input");
                    port_type = 'I';
                    break;
                case 0x03:
                    sprintf(stemp, "bidirectional");
                    port_type = 'B';
                    break;           
            }            
            if (opt_list) output_string(stemp);
            ptr++;
            
            /* Print collected information for kicad */
            if (opt_kicad) {
                sprintf(stemp, "\nF%d \"", i+1); /* Field number starts with 2 */
                output_string(stemp);
            }
            if (opt_list) {
                sprintf(stemp, "\nSheet net %d port name: ", i);
                output_string(stemp);
            }
            temp = 0; memcpy(&temp, &input_buffer[ptr], 1);
            ptr++;
            memcpy(&stemp, &input_buffer[ptr], temp);
            stemp[temp] = '\0';
            strncpy(port_name, stemp, MAXSTRING - 1);
            if (opt_kicad) output_string(stemp);
            ptr = ptr + temp;
            
            if (opt_kicad) {
                if (port_side == 'L') {
                    port_x = pos_x_top; 
                    port_y = pos_y_top + port_offset;
                } else {
                    port_x = pos_x_top + width; 
                    port_y = pos_y_top + port_offset;
                }
                sprintf(stemp, "\" %c %c %d %d 60", port_type, port_side, port_x, port_y);
                output_string(stemp);
            } else if (opt_geda) {
               if (port_side == 'L') {
                    port_x = pos_x_top; 
                    port_y = pos_y_top - port_offset;
                } else {
                    port_x = pos_x_top + width; 
                    port_y = pos_y_top - port_offset;
                }
                /* create sheet net ports as part of the sheet component */
                geda_sheet_port(port_x, port_y, i, port_name, port_side, port_type);
            }
    }
    if (opt_kicad) {
        sprintf(stemp, "\n$EndSheet");
        output_string(stemp);
    } else if (opt_geda) {
        /* end of embedded component,
         * add sheet name, filename
         */
        sprintf(stemp, "\n]\n{");
        output_string(stemp);
        sprintf(stemp, "\nT %d %d 8 6 1 1 0 0 1", pos_x_top, pos_y_top + 20);
        output_string(stemp);
        sprintf(stemp, "\npartno=%s", sheet_name);
        output_string(stemp);
        sprintf(stemp, "\nT %d %d 8 6 1 1 0 2 1", pos_x_top, pos_y_top - height - 10);
        output_string(stemp);
        sprintf(stemp, "\ndevice=%s", sheet_filename);
        output_string(stemp);
        sprintf(stemp, "\nT %d %d 8 6 0 0 0 2 1", pos_x_top, pos_y_top - height - 60);
        output_string(stemp);
        sprintf(stemp, "\nsource=%s", sheet_filename);
        output_string(stemp);
        sprintf(stemp, "\n}");
        output_string(stemp);
    }
    if (opt_list) { 
        sprintf(stemp, "\n");
        output_string(stemp);
    }
}

/* process_component (Tag 02)
 * 
 */
void process_component(void)
{
    int byte_count = 0;

    unsigned char stemp[MAXSTRING];
    int temp;
    int ptr;
    int flags, attr_vis, mask_vis, sheetpart, i; /* attr_vis: all attribute visibility flags */
    unsigned char part[MAXSTRING];
    unsigned char part_value[MAXSTRING];
    unsigned char ref[MAXSTRING];
    unsigned char attr[MAXSTRING];
    unsigned char datasheet[MAXSTRING];
    char *letters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";
    int comp_x, comp_y;
    int ref_offx, ref_offy;
    int part_offx, part_offy;
    int attr_offx, attr_offy;
    int sheetpart_offx, sheetpart_offy;
    int dev_num, dev_equiv;
    int am, bm, cm, dm; /* mirroring matrix NOTE: NOT USED */
    int ax, bx, cx, dx; /* orientation matrix NOTE: NOT USED*/
    int a, b, c, d; /* component orientation */ 
    int ref_inv, part_inv, attr_inv, sheetpart_inv; /* individual visibility flags */
    int uniq_1, uniq_2; /* Unique component identifier */
    int gangle, gmirror; /* geda: angle and mirroring */
    
    ptr = iptr + 2; /* Initialize local pointer - skip byte count */

    memcpy(&byte_count, &input_buffer[iptr], 2);
    iptr = iptr + byte_count + 2;
    infileptr = infileptr + byte_count + 3;
    total_components++;
    
    if (opt_verbose) {
        sprintf(stemp, "\n#ptr: %X", ptr);
        output_string(stemp);
    }

    temp = 0; memcpy(&temp, &input_buffer[components], 1);
    sprintf(stemp, "\nComponent part: ");
    if (opt_list) output_string(stemp);
    components++;
    memcpy(&stemp, &input_buffer[components], temp);
    stemp[temp] = '\0';
    if (opt_list) output_string(stemp);
    strncpy(part, stemp, MAXSTRING);
    components = components + temp;
    
    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    if (temp > 32767) temp -= 65535;
    sprintf(stemp, "\nComponent position x: %d", temp);
    if (opt_list) output_string(stemp);
    comp_x = temp * SCALE + OFF_X;
    ptr += 2;

    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    if (temp > 32767) temp -= 65535;
    sprintf(stemp, "\nComponent position y: %d", temp);
    if (opt_list) output_string(stemp);
    comp_y = temp * SCALE + OFF_Y;
    if (opt_geda) comp_y = geda_y - comp_y;
    ptr += 2;

    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    if (temp > 32767) temp -= 65535;
    sprintf(stemp, "\nComponent reference designator offset x: %d", temp);
    if (opt_list) output_string(stemp);
    ref_offx = temp * SCALE;
    ptr += 2;
    
    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    if (temp > 32767) temp -= 65535;
    sprintf(stemp, "\nComponent reference designator offset y: %d", temp);
    if (opt_list) output_string(stemp);
    ref_offy = temp * SCALE;
    ptr += 2;    

    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    if (temp > 32767) temp -= 65535;
    sprintf(stemp, "\nComponent part value offset x: %d", temp);
    if (opt_list) output_string(stemp);
    part_offx = temp * SCALE;
    ptr += 2;
    
    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    if (temp > 32767) temp -= 65535;
    sprintf(stemp, "\nComponent part value offset y: %d", temp);
    if (opt_list) output_string(stemp);
    part_offy = temp * SCALE;
    ptr += 2;
    
    temp = 0; memcpy(&temp, &input_buffer[ptr], 1);
    dev_num = temp & 0x1f;
    sprintf(stemp, "\nComponent device number: %d", dev_num);
    if (opt_list) output_string(stemp);
    ptr += 1;
    
    subspaces(part);    /* NOTE: kicad: spaces not allowed in part names */
    get_part_data(part);
    
    /* orcad: mirror is around the y axis 
     * Identity matrix [1 0 0 1]
     * Transformation matrix around y is [am bm cm dm] = [-1 0 0 1] (around x is [1 0 0 -1])
     * [x]   [a b]   [ax + by]
     * | | x |   | = |       ]
     * [y]   [c d]   [cx + dy]
     * NOTE: transformation matrices are not used in this version (too much weirdness).
     * 
     * kicad: the following transformations are used (X, Y are component width, height):
     * 
     * Orientation      Transformation  X offset    Y offset
     * -----------      --------------  --------    --------
     * Right             1  0  0 -1     0           0
     * Right, mirrored  -1  0  0 -1     X           0
     * Up                0 -1 -1  0     0           X
     * Up, mirrored      0 -1  1  0     Y           X   (marked as down, mirrored in orcad!)
     * Down              0  1  1  0     Y           0
     * Down, mirrored    0  1 -1  0     0           0   (marked as up, mirrored in orcad!)
     * Over             -1  0  0  1     X           Y
     * Over, mirrored    1  0  0  1     0           Y
     * 
     */
    sprintf(stemp, "\nComponent device orientation (%x) : ", temp);
    if (opt_list) output_string(stemp);
    switch (temp & 0x60) {
        case 0x00:
            sprintf(stemp, "right");
//            ax=1; bx=0; cx=0; dx=1; /* 0 degrees */
            if (temp &0x80) {   /* check mirrored flag */
                a=-1; b=0; c=0; d=-1;
                comp_x = comp_x + x_size;
                gangle = 0; gmirror = 1;
            } else {
                a=1; b=0; c=0; d=-1;
                gangle = 0; gmirror = 0;
            }
            break;
        case 0x20:
            sprintf(stemp, "up");
  //          ax=0; bx=1; cx=-1; dx=0; /* +90 degrees) */
            if (temp &0x80) {
                a=0; b=-1; c=1; d=0;
                gangle = 90; gmirror = 1;
            } else {
                a=0; b=-1; c=-1; d=0;
                if (opt_kicad) comp_y = comp_y + x_size;
                else if (opt_geda) comp_y = comp_y - x_size;
                gangle = 90; gmirror = 0;
            }
            break;
        case 0x40:
            sprintf(stemp, "over");
  //          ax=-1; bx=0; cx=0; dx=-1; /* 180 degrees */
            if (temp &0x80) {
                a=1; b=0; c=0; d=1;
                if (opt_kicad) comp_y = comp_y + y_size;
                else if (opt_geda) comp_y = comp_y - y_size;
                gangle = 180; gmirror = 1;
            } else {
                a=-1; b=0; c=0; d=1;
                comp_x = comp_x + x_size;
                if (opt_kicad) comp_y = comp_y + y_size;
                else if (opt_geda) comp_y = comp_y - y_size;
                gangle = 180; gmirror = 0;
            }
            break;
        case 0x60:
            sprintf(stemp, "down");
  //          ax=0; bx=-1; cx=1; dx=0; /* 270 degrees) */
            if (temp &0x80) {
                a=0; b=1; c=-1; d=0;
                comp_x = comp_x + y_size;
                if (opt_kicad) comp_y = comp_y + x_size;
                else if (opt_geda) comp_y = comp_y - x_size;
                gangle = 270; gmirror = 1;
            } else {
                a=0; b=1; c=1; d=0;
                comp_x = comp_x + y_size;
                gangle = 270; gmirror = 0;
            }
            break;
    }
    if (opt_list) output_string(stemp);

    if (temp &0x80) {
        sprintf(stemp, ", mirrored ");
        if (opt_list) output_string(stemp);
//         am=-1; bm=0; cm=0; dm=1; /* mirror around y */
//         a = am*ax+bm*cx;
//         b = am*bx+bm*dx;
//         c = cm*ax+dm*cx;
//         d = cm*bx+dm*dx;
//         ax = a; bx = b; cx = c; dx = d; /* modified transformation matrix */
    }
    //am=1; bm=0; cm=0; dm=1; /* identity matrix if not mirrored */
    //am=1; bm=0; cm=0; dm=-1; /* kicad: mirror about x appears to be the default !!
    /* Calculate device orientation */
    
//     a = am*ax+bm*cx;
//     b = am*bx+bm*dx;
//     c = cm*ax+dm*cx;
//     d = cm*bx+dm*dx;
    
    temp = 0; memcpy(&temp, &input_buffer[ptr], 1);
    sprintf(stemp, "\nComponent flags (hex): %2.2x", temp);
    if (opt_list) output_string(stemp);
    ptr += 1;
    flags = temp;
    if (flags & 0x02) ref_inv = 1; else ref_inv = 0;
    if (flags & 0x04) part_inv = 1; else part_inv = 0;
    
    dev_equiv = 1; /* kicad - this is the normal symbol */
    if (flags & 0x1) {
        sprintf(stemp, "\nComponent uses DeMorgan equivalent symbol");
        if (opt_list) output_string(stemp);
        dev_equiv = 2; /* kicad - this is the equivalent symbol (1 is normal) */
    }
    
    temp = 0; memcpy(&temp, &input_buffer[ptr], 1);
    sprintf(stemp, "\nComponent attribute visibility (hex): %2.2x", temp);
    if (opt_list) output_string(stemp);
    ptr += 1;
    attr_vis = temp;

    temp = 0; memcpy(&temp, &input_buffer[ptr], 1);
    sprintf(stemp, "\nComponent sheet part flags (hex): %2.2x", temp);
    if (opt_list) output_string(stemp);
    ptr += 1;
    sheetpart = temp;
    if (version == 1) sheetpart = 0; /* NOTE: not used for older version */
    
    /* Update and unique component (time stamp) information */
    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nComponent information separator (hex): %2.2x", temp);
    if (opt_list) output_string(stemp);
    ptr += 2;


    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nComponent update cycle: %d ", temp);
    if (opt_list) output_string(stemp);
    ptr += 2;    

    if (version == 2) { /* NOTE: skipped if older version */
    temp = 0; memcpy(&temp, &input_buffer[ptr], 2); /* follows update cycle information */
    sprintf(stemp, "(hex): %2.2x", temp);
    if (opt_list) output_string(stemp);
    ptr += 2;
    }

    /* Orcad SDT unique component identification (uses a time stamp with 50mS increments)
     * The LSB is incremented for each component 
     * within the same update cycle.
     */
    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nComponent identifier -- time stamp (hex): %2.2x -- ", temp);
    if (opt_list) output_string(stemp);
    ptr += 2;
    
    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "%2.2x ", temp);
    if (opt_list) output_string(stemp);
    uniq_2 = temp;
    ptr += 2;

    if (version == 2) { /* NOTE: skipped if older version */
    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "%2.2x ", temp);
    if (opt_list) output_string(stemp);
    uniq_1 = temp;
    ptr += 2;
    }


    sprintf(stemp, "\nComponent reference: ");
    if (opt_list) output_string(stemp);
    temp = 0; memcpy(&temp, &input_buffer[ptr], 1);
    ptr++;
    memcpy(&stemp, &input_buffer[ptr], temp);
    stemp[temp] = '\0';
    if (opt_list) output_string(stemp);
    strncpy(ref, stemp, MAXSTRING);
    ptr = ptr + temp;
    
    sprintf(stemp, "\nComponent part value: ");
    if (opt_list) output_string(stemp);
    temp = 0; memcpy(&temp, &input_buffer[ptr], 1);
    ptr++;
    memcpy(&stemp, &input_buffer[ptr], temp);
    stemp[temp] = '\0';
    if (opt_list) output_string(stemp);
    strncpy(part_value, stemp, MAXSTRING);
    ptr = ptr + temp;
    
    /* Combine information for kicad component records
     */
    int ref_pos_x, ref_pos_y;
    int part_value_pos_x, part_value_pos_y;
    if (opt_kicad) {
         /* Default field placement 
         * (should correspond to the orcad_lib_convert algorithm) 
         */
        if (x_size > 0 && x_size <= 200) {
            ref_pos_x = x_size + 40;
            ref_pos_y = -(y_size / 2) + 10;  /* Bottom alignment */
            part_value_pos_x = x_size + 40;
            part_value_pos_y = -(y_size / 2) - 10; /* Top alignment */
        } else {
            ref_pos_x = 10;
            ref_pos_y = 15;
            part_value_pos_x = 10;
            part_value_pos_y = -y_size - 15;
        }
        //FIXME: zero out reference and part number offsets
            ref_offx = 0; ref_offy = 0;
            part_offx = 0; part_offy = 0;
        // FIXME: DEBUG
//         if (part_offx || part_offy || ref_offx || ref_offy) {
//             printf("\n%s %d %d %d %d", part, ref_offx, ref_offy, part_offx, part_offy);
//         }
        /* kicad: spaces not allowed in part number, reference
         * substituded by underscore
         */
        subspaces(part); /* NOTE: REDUNDANT */
        subspaces(ref);
        subquotes(part_value);
        /* 
         * Remove part designator which will be re-inserted by kicad 
         */
        int reflast = strlen(ref)?strlen(ref)-1:0;
        if (reflast && (dev_num > 0) && (dev_num <= 26) && (ref[reflast] == letters[dev_num - 1])) {
            ref[reflast] = '\0';
        }
        sprintf(stemp, "\n$Comp\nL %s%c%s %s", (strlen(part_library)?(char *)part_library:" "), (strlen(part_library)?':':' '), part, (strlen(ref)?(char *)ref:"~"));
        output_string(stemp);
        sprintf(stemp, "\nU %d %d %4.4X%4.4X", dev_num, dev_equiv, uniq_1, uniq_2); /* Unique component identifier */
        output_string(stemp);
        sprintf(stemp, "\nP %d %d", comp_x, comp_y);
        output_string(stemp);
        sprintf(stemp, "\nF 0 \"%s\" H %d %d 60 000%1d L BNN", (strlen(ref)?(char *)ref:"~"), comp_x + ref_pos_x + ref_offx, comp_y + ref_pos_y + ref_offy, ref_inv); /* reference */
        output_string(stemp);
        sprintf(stemp, "\nF 1 \"%s\" H %d %d 60 000%1d L TNN", part_value, comp_x + part_value_pos_x + part_offx, comp_y + part_value_pos_y + part_offy, part_inv); /* part value */
        output_string(stemp);
        sprintf(stemp, "\nF 2 \"\" H %d %d 60 0001 C CNN", comp_x, comp_y); /* footprint NOTE: placeholder - currently invisible */
        output_string(stemp);
        sprintf(stemp, "\nF 3 \"\" H %d %d 60 0001 C CNN", comp_x, comp_y); /* datasheet NOTE: placeholder - currently invisible */
        output_string(stemp);
    } else if (opt_geda) {
        if (x_size > 0 && x_size <= 200) {
            ref_pos_x = x_size + 40;
            ref_pos_y = -(y_size / 2) + 10;  /* Bottom alignment */
            part_value_pos_x = x_size + 40;
            part_value_pos_y = -(y_size / 2) - 10; /* Top alignment */
        } else {
            ref_pos_x = 10;
            ref_pos_y = 15;
            part_value_pos_x = 10;
            part_value_pos_y = -y_size - 15;
        }
        //FIXME: zero out reference and part number offsets
        ref_offx = 0; ref_offy = 0;
        part_offx = 0; part_offy = 0;
        /* C x y selectable angle mirror basename
         */
        sprintf(stemp,"\nC %d %d 1 %d %d %s-%d.sym", \
            comp_x, comp_y, gangle, gmirror, part, dev_equiv);
        output_string(stemp);
        sprintf(stemp,"\n{");
        output_string(stemp);
        sprintf(stemp, "\nT %d %d 5 5 0 0 0 0 1", \
            comp_x + part_value_pos_x + part_offx, comp_y + part_value_pos_y + part_offy, part_inv?0:1);
        output_string(stemp);
        sprintf(stemp,"\ndevice=%s", part);
        output_string(stemp);
        if (strlen(ref)) {
            sprintf(stemp, "\nT %d %d 5 5 %d 1 0 0 1", \
                comp_x + ref_pos_x + ref_offx, comp_y + ref_pos_y + ref_offy, ref_inv?0:1);
            output_string(stemp);
            sprintf(stemp,"\nrefdes=%s", ref);
            output_string(stemp);
        }
        if (dev_num > 1) {
            sprintf(stemp, "\nT %d %d 5 5 0 0 0 0 1", \
                comp_x + ref_pos_x + ref_offx + 100, comp_y + ref_pos_y + ref_offy);
            output_string(stemp);
            sprintf(stemp,"\nslot=%d", dev_num);
            output_string(stemp);
        }
        if (strlen(part_value)) {
            sprintf(stemp, "\nT %d %d 5 5 %d 1 0 2 1", \
                comp_x + part_value_pos_x + part_offx, comp_y + part_value_pos_y + part_offy, part_inv?0:1);
            output_string(stemp);
            sprintf(stemp,"\nvalue=%s", part_value);
            output_string(stemp);
        }
        sprintf(stemp,"\n}");
        output_string(stemp);

    }
    
    
    /* If the "Include field attributes" flag is set,
     * get attributes.
     */
    if (flags & 0x40) {
        for (i=1; i<=8; i++) {

            /* Attribute visibility */
            if (i == 1) mask_vis = 1; /* initialize attribute visibility mask */
            else mask_vis = mask_vis << 1;
            if (attr_vis & mask_vis) attr_inv = 1; else attr_inv = 0;

            /* Skip empty (zero length) attributes -- 
             * currently pointing at first byte of xoffset  
             */
            temp = 0; memcpy(&temp, &input_buffer[ptr+4], 1);
            if (temp == 0) {
                ptr += 5; /* Skip over the empty attribute */
                continue;
            }

            temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
            if (temp > 32767) temp -= 65535;
            sprintf(stemp, "\nComponent attribute %d offset x: %d", i, temp);
            if (opt_list) output_string(stemp);
            attr_offx = temp * SCALE;
            ptr += 2;
            
            temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
            if (temp > 32767) temp -= 65535;
            sprintf(stemp, "\nComponent attribute %d offset y: %d", i, temp);
            if (opt_list) output_string(stemp);
            attr_offy = temp * SCALE;
            ptr += 2;
            
            sprintf(stemp, "\nComponent attribute %d: ", i);
            if (opt_list) output_string(stemp);
            temp = 0; memcpy(&temp, &input_buffer[ptr], 1);
            ptr++;
            memcpy(&stemp, &input_buffer[ptr], temp);
            stemp[temp] = '\0';
            if (opt_list) output_string(stemp);
            strncpy(attr, stemp, MAXSTRING);
            ptr = ptr + temp;
            
            if(opt_kicad) {
                /* attributes start at F 5 */
                sprintf(stemp, "\nF %d \"%s\" H %d %d 60 000%1d C CNN", \
                    i+4, attr, comp_x + attr_offx, comp_y + attr_offy, attr_inv); 
                output_string(stemp);
            } else if (opt_geda) {
                /* only the value of an attribute is shown, if visible */
                sprintf(stemp, "\nT %d %d 5 3 %d 1 0 0 1", \
                comp_x + attr_offx, comp_y + attr_offy, attr_inv?0:1);
                output_string(stemp);
                sprintf(stemp,"\nattribute%1d=%s", i, attr);
                output_string(stemp); 
            }
        }
    }

    /* If the "Include sheetpart name" flag is set,
     * get sheetpart field.
     */
    if (sheetpart & 0x80) {
    
        if (sheetpart & 0x40) sheetpart_inv = 1; else sheetpart_inv = 0;

        temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
        if (temp > 32767) temp -= 65535;
        sprintf(stemp, "\nComponent sheetpart offset x: %d", temp);
        if (opt_list) output_string(stemp);
        sheetpart_offx = temp * SCALE;
        ptr += 2;
        
        temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
        if (temp > 32767) temp -= 65535;
        sprintf(stemp, "\nComponent sheetpart offset y: %d", temp);
        if (opt_list) output_string(stemp);
        sheetpart_offy = temp * SCALE;
        ptr += 2;
        
        sprintf(stemp, "\nComponent sheetpart: ", i);
        if (opt_list) output_string(stemp);
        temp = 0; memcpy(&temp, &input_buffer[ptr], 1);
        ptr++;
        memcpy(&stemp, &input_buffer[ptr], temp);
        stemp[temp] = '\0';
        if (opt_list) output_string(stemp);
        strncpy(datasheet, stemp, MAXSTRING);
        ptr = ptr + temp;
        
        if(opt_kicad) {
            /* datasheet is F 3 */
            sprintf(stemp, "\nF 3 \"%s\" H %d %d 60 000%1d C CNN", \
                datasheet, comp_x + sheetpart_offx, comp_y + sheetpart_offy, sheetpart_inv); 
            output_string(stemp);
        } else if (opt_geda) {
                sprintf(stemp, "\nT %d %d 5 3 %d 0 0 0 1", \
                comp_x + sheetpart_offx, comp_y + sheetpart_offy, sheetpart_inv?0:1);
                output_string(stemp);
                sprintf(stemp,"\nsheetpart=%s", datasheet);
                output_string(stemp); 
        }
    }
    if(opt_kicad) {
        sprintf(stemp, "\n    %d %d %d", dev_num, comp_x, comp_y);
        output_string(stemp);
        sprintf(stemp, "\n    %d %d %d %d", a, b, c, d);
        output_string(stemp);
        sprintf(stemp, "\n$EndComp");
        output_string(stemp);
    }
    if (opt_list) { 
        sprintf(stemp, "\n");
        output_string(stemp);
    }
}

/* process_wire (Tag 03)
 * 
 */
void process_wire(void)
{
    int byte_count = 0;
    
    unsigned char stemp[MAXSTRING];
    int temp;
    int ptr;
    int wire_start_x, wire_start_y, wire_end_x, wire_end_y;
    
    ptr = iptr + 2; /* Initialize local pointer - skip byte count */
    
    memcpy(&byte_count, &input_buffer[iptr], 2);
    iptr = iptr + byte_count + 2;
    infileptr = infileptr + byte_count + 3;
    total_wires++;

    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nWire start x: %d", temp);
    if (opt_list) output_string(stemp);
    wire_start_x = temp * SCALE + OFF_X;
    ptr += 2;

    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nWire start y: %d", temp);
    if (opt_list) output_string(stemp);
    wire_start_y = temp * SCALE + OFF_Y;
    if (opt_geda) wire_start_y = geda_y - wire_start_y;
    ptr += 2;

    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nWire end x: %d", temp);
    if (opt_list) output_string(stemp);
    wire_end_x = temp * SCALE + OFF_X;
    ptr += 2;

    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nWire end y: %d", temp);
    if (opt_list) output_string(stemp);
    wire_end_y = temp * SCALE + OFF_Y;
    if (opt_geda) wire_end_y = geda_y - wire_end_y;
    ptr += 2;
    
    if (opt_kicad) {
        sprintf(stemp, "\nWire Wire Line\n    %d %d %d %d", wire_start_x, wire_start_y, wire_end_x, wire_end_y);
        output_string(stemp);
    } else if (opt_geda) {
        /* geda: wires are nets */
        sprintf(stemp, "\nN %d %d %d %d 4", wire_start_x, wire_start_y, wire_end_x, wire_end_y);
        output_string(stemp);
    }

    if (opt_list) { 
        sprintf(stemp, "\n");
        output_string(stemp);
    }
}

/* process_bus (Tag 04)
 * 
 */
void process_bus(void)
{
    int byte_count = 0;

    unsigned char stemp[MAXSTRING];
    int temp;
    int ptr;
    int bus_start_x, bus_start_y, bus_end_x, bus_end_y;
    
    ptr = iptr + 2; /* Initialize local pointer - skip byte count */
    
    memcpy(&byte_count, &input_buffer[iptr], 2);
    iptr = iptr + byte_count + 2;
    infileptr = infileptr + byte_count + 3;
    total_busses++;
    
    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nBus start x: %d", temp);
    if (opt_list) output_string(stemp);
    bus_start_x = temp * SCALE + OFF_X;
    ptr += 2;

    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nBus start y: %d", temp);
    if (opt_list) output_string(stemp);
    bus_start_y = temp * SCALE + OFF_Y;
    if (opt_geda) bus_start_y = geda_y - bus_start_y;
    ptr += 2;

    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nBus end x: %d", temp);
    if (opt_list) output_string(stemp);
    bus_end_x = temp * SCALE + OFF_X;
    ptr += 2;

    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nBus end y: %d", temp);
    if (opt_list) output_string(stemp);
    bus_end_y = temp * SCALE + OFF_Y;
    if (opt_geda) bus_end_y = geda_y - bus_end_y;
    ptr += 2;
    
    if (opt_kicad) {
        sprintf(stemp, "\nWire Bus Line\n    %d %d %d %d", bus_start_x, bus_start_y, bus_end_x, bus_end_y);
        output_string(stemp);
    } else if (opt_geda) {
        /* geda: connections to the end of the bus do not work
         * geda: ends will be lengthened by 10 mils to accommodate
         */
        if (bus_start_y == bus_end_y) {
            bus_start_x = (bus_start_x < bus_end_x)? bus_start_x - 10: bus_start_x + 10;
            bus_end_x = (bus_start_x < bus_end_x)? bus_end_x + 10: bus_end_x - 10;
        } else if (bus_start_x == bus_end_x) {
            bus_start_y = (bus_start_y < bus_end_y)? bus_start_y - 10: bus_start_y + 10;
            bus_end_y = (bus_start_y < bus_end_y)? bus_end_y + 10: bus_end_y - 10;
        }
        sprintf(stemp, "\nU %d %d %d %d 10 0", bus_start_x, bus_start_y, bus_end_x, bus_end_y);
        output_string(stemp);
    }
    
    if (opt_list) { 
        sprintf(stemp, "\n");
        output_string(stemp);
    }
}

/* process_junction (Tag 05)
 * 
 */
void process_junction(void)
{
    int byte_count = 0;

    unsigned char stemp[MAXSTRING];
    int temp;
    int ptr;
    int junction_x, junction_y;
    
    ptr = iptr + 2; /* Initialize local pointer - skip byte count */
    
    memcpy(&byte_count, &input_buffer[iptr], 2);
    iptr = iptr + byte_count + 2;
    infileptr = infileptr + byte_count + 3;
    total_junctions++;

    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nJunction x: %d", temp);
    if (opt_list) output_string(stemp);
    junction_x = temp * SCALE + OFF_X;
    ptr += 2;

    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nJunction y: %d", temp);
    if (opt_list) output_string(stemp);
    junction_y = temp * SCALE + OFF_Y;
    if (opt_geda) junction_y = geda_y - junction_y;
    ptr += 2;
    
    if (opt_kicad) {
        sprintf(stemp, "\nConnection ~ %d %d", junction_x, junction_y);
        output_string(stemp);
    } else if (opt_geda) {
        /* this appears to be redundant as gEDA does not have a junction item FIXME: CHECK*/
//         sprintf(stemp, "\nN %d %d %d %d 4", junction_x, junction_y, junction_x, junction_y);
//         output_string(stemp);
    }
    
    if (opt_list) { 
        sprintf(stemp, "\n");
        output_string(stemp);
    }
}

/* process_module_port (Tag 06)
 * 
 * orcad: module_x and module_y is the leftmost position of the module label
 * 
 * kicad: module_x and module_y is the connection point.
 * kicad: There is no module width correspondence in kicad
 * kicad: Shape is determined by module port type.
 * 
 * geda:  module is an embedded component with zero-length pins
 *        at both ends, with the same netname
 *
 */
void process_module_port(void)
{
    int byte_count = 0;

    unsigned char stemp[MAXSTRING];
    unsigned char module[MAXSTRING];
    unsigned char module_type[MAXSTRING];
    int temp;
    int ptr;
    int module_x, module_y, module_size, module_conn_x; 
    int module_function, module_shape, module_orientation, module_width;
    
    ptr = iptr + 2; /* Initialize local pointer - skip byte count */
    
    memcpy(&byte_count, &input_buffer[iptr], 2);
    iptr = iptr + byte_count + 2;
    infileptr = infileptr + byte_count + 3;
    total_module_ports++;
    
    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    if (opt_list) {
        sprintf(stemp, "\nModule port x: %d", temp);
        output_string(stemp);
    }
    module_x = temp * SCALE + OFF_X;
    ptr += 2;

    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    if (opt_list) {
        sprintf(stemp, "\nModule port y: %d", temp);
        output_string(stemp);
    }
    module_y = temp * SCALE + OFF_Y;
     if (opt_geda) module_y = geda_y - module_y;
    ptr += 2;
    
    temp = 0; memcpy(&temp, &input_buffer[ptr], 1);
    if (opt_list) {
        sprintf(stemp, "\nModule port width (pixels): %d", temp);
        output_string(stemp);
    }
    module_width = temp * SCALE;
    ptr += 1;
    
    temp = 0; memcpy(&temp, &input_buffer[ptr], 1);
    if (opt_list) {
        sprintf(stemp, "\nModule port type: ");
        output_string(stemp);
    }
    ptr += 1;
    
    module_function = UNSPECIFIED;

    switch (temp & 0xc0) {
        case 0x00:
            sprintf(stemp, "unspecified");
            if (opt_kicad) sprintf(module_type,"UnSpc");
            else if (opt_geda) sprintf(module_type,"pas");
            module_function = UNSPECIFIED;
            break;
        case 0x40:
            sprintf(stemp, "output");
            if (opt_kicad) sprintf(module_type,"Output");
            else if (opt_geda) sprintf(module_type,"out");
            module_function = OUTPUT;
            break;
        case 0x80:
            sprintf(stemp, "input");
            if (opt_kicad) sprintf(module_type,"Input");
            else if (opt_geda) sprintf(module_type,"in");
            module_function = INPUT;
            break;
        case 0xc0:
            sprintf(stemp, "bidirectional");
            if (opt_kicad) sprintf(module_type,"BiDi");
            else if (opt_geda) sprintf(module_type,"io");
            module_function = BIDIRECTIONAL;
            break;           
    }
    if (opt_list) output_string(stemp);
    
    /* At this point we are re-using the contents of temp 
     */
    if (opt_list) {
        sprintf(stemp, "\nModule port shape: ", temp);
        output_string(stemp);
    }
    module_orientation = 0; /* kicad: default - from left */
    switch (temp & 0x30) {
        case 0x00:
            sprintf(stemp, "neither pointing");
            module_shape = NEITHER;
            module_conn_x = module_x + module_width; /* connection point for kicad */
            break;
        case 0x10:
            sprintf(stemp, "left pointing");
            module_shape = LEFT;
            module_width += 100;
            if (module_function == INPUT) {
                module_conn_x = module_x + module_width; /* connection point for kicad */
            } else if (module_function == OUTPUT) {
                module_orientation = 0; /* kicad: output to the left */
                module_conn_x = module_x + module_width; /* connection point for kicad */
            }            
            break;
        case 0x20:
            sprintf(stemp, "right pointing");
            module_shape = RIGHT;
            module_width += 100;
            if (module_function == INPUT) {
                module_conn_x = module_x + module_width; /* connection point for kicad */
            } else if (module_function == OUTPUT) {
                module_orientation = 2; /* kicad: to the right */
                module_conn_x = module_x + module_width; /* connection point for kicad */
            }
            break;
        case 0x30:
            sprintf(stemp, "both pointing"); /* right and left arrow widths are added to module width */
            module_shape = BOTH; /* orient both pointing to the left */
            module_width += 200;
            module_conn_x = module_x + module_width; /* connection point for kicad */
            break;           
    }
    if (opt_list) output_string(stemp);
    
    if (opt_list) {
        sprintf(stemp, "\nModule port label: ", temp);
        output_string(stemp);
    }
    temp = 0; memcpy(&temp, &input_buffer[ptr], 1);
    ptr++;
    memcpy(&stemp, &input_buffer[ptr], temp);
    stemp[temp] = '\0';
    if (opt_list) output_string(stemp);
    strncpy(module, stemp, MAXSTRING);
    ptr = ptr + temp;
    /* 
     * orcad: module is always horizontal and
     *        position (x,y) is at the left edge of the module port
     * 
     * kicad: output module, connection point is at the flat part
     *        input module, connection is the arrow point
     *        bidirectional, tristate, passive modules, 
     *        connection at right  (left orientation) or left (right orientation)
     * 
     * geda:  an embedded module component is created
     * 
     * HACK -- in order to assure module connectivity
     * kicad: a wire connects both sides of the module port to assure connectivity (-m)
     * kicad: or geda: a named node/net (-n) is created in place of the module 
     * geda:  connection points (pins) at both ends of the module with the same "netname=" attribute
     */
    if (opt_kicad) {
        if (opt_nodes || opt_modules) {
            sprintf(stemp, "\nWire Wire Line\n    %d %d %d %d",\
                module_x, module_y, module_x + module_width + ((version ==2)?0:100), module_y); /* version 1 needs an extra 100 mils... */
            output_string(stemp);
        } 
        if (opt_nodes) {
            sprintf(stemp, "\nText Label %d %d 0 60 ~\n%s", \
            module_conn_x, module_y, module);
            output_string(stemp);
        } else {
            sprintf(stemp, "\nText Glabel %d %d %d 60 %s ~ 0\n%s", \
            module_conn_x, module_y, module_orientation, module_type, module);
            output_string(stemp);
        }
    } else if (opt_geda) {
        if (opt_nodes) {
            /* Create a net, the length of a module port */
            sprintf(stemp, "\nN %d %d %d %d 4",  module_x, module_y, module_x + module_width, module_y);
            output_string(stemp);
            sprintf(stemp, "\n{\nT %d %d 5 6 1 1 0 0 1\nnetname=%s\n}",module_x, module_y, module); 
            output_string(stemp);
        } else {
            /* opt_modules is the default for geda */
            /* Create an embedded module component 
             * origin is lower left corner
             * connection points (pins) are at middle left and right - both ends have the same netname
             */
            sprintf(stemp, "\nC %d %d 1 0 0 EMBEDDED_module_port_%4.4X-1.sym\n[", module_x, module_y - 150, total_module_ports);
            output_string(stemp);
            sprintf(stemp, "\nP %d %d %d %d 1 0 0\n{", module_x, module_y, module_x, module_y);
            output_string(stemp);
            sprintf(stemp, "\nT %d %d 5 6 0 1 0 1 1\npinseq=1", module_x, module_y);
            output_string(stemp);
            sprintf(stemp, "\nT %d %d 5 6 0 1 0 1 1\npinnumber=1", module_x + 100, module_y);
            output_string(stemp);
            sprintf(stemp, "\nT %d %d 5 6 1 1 0 1 1\nnetname=%s", module_x + 120, module_y, module);
            output_string(stemp);
            sprintf(stemp, "\nT %d %d 5 6 0 1 0 1 1\npintype=%s\n}", module_x + 300, module_y, module_type);
            output_string(stemp);
            sprintf(stemp, "\nP %d %d %d %d 1 0 0\n{", module_x + module_width, module_y, module_x + module_width, module_y);
            output_string(stemp);
            sprintf(stemp, "\nT %d %d 5 6 0 1 0 1 1\npinseq=2", module_x, module_y);
            output_string(stemp);
            sprintf(stemp, "\nT %d %d 5 6 0 1 0 1 1\npinnumber=2", module_x + 100, module_y);
            output_string(stemp);
            sprintf(stemp, "\nT %d %d 5 6 0 1 0 1 1\nnetname=%s", module_x + 220, module_y, module);
            output_string(stemp);
            sprintf(stemp, "\nT %d %d 5 6 0 1 0 1 1\npintype=%s\n}", module_x + 300, module_y, module_type);
            output_string(stemp);

            /* module outline created using a 'path' directive 
             */
            sprintf(stemp,"\nH 3 10 0 0 -1 -1 0 -1 -1 -1 -1 -1 3");
            output_string(stemp);
            sprintf(stemp,"\nM %d %d", module_x, module_y);
            output_string(stemp);
            
            /* four variants - neither pointing, left pointing, right pointing, both pointing 
             * using relative coordinates
             */
            switch (module_shape) {
                case NEITHER:
                    sprintf(stemp,"\nl %d,%d %d,%d %d,%d %d,%d", \
                        0,50, module_width,0, 0,-100, -module_width,0);
                    output_string(stemp);
                    break;
                case LEFT:
                    sprintf(stemp,"\nl %d,%d %d,%d %d,%d %d,%d", \
                        100,50, module_width-100,0, 0,-100, -module_width+100,0);
                    output_string(stemp);
                    break;
                case RIGHT:
                    sprintf(stemp,"\nl %d,%d %d,%d %d,%d %d,%d %d,%d", \
                        0,50, module_width-100,0, 100,-50, -100,-50, -module_width+100,0);
                    output_string(stemp);
                    break;
                case BOTH:
                    sprintf(stemp,"\nl %d,%d %d,%d %d,%d %d,%d %d,%d", \
                        100,50, module_width-200,0, 100,-50, -100,-50, -module_width+200,0);
                    output_string(stemp);
                    break;
            }
            sprintf(stemp,"\nz\n]"); /* Z or z closes the path */
            output_string(stemp);           
        }
    }
    if (opt_list) { 
        sprintf(stemp, "\n");
        output_string(stemp);
    }
}

/* process_label (Tag 07)
 * 
 */
void process_label(void)
{
    int byte_count = 0;

    unsigned char stemp[MAXSTRING];
    unsigned char label[MAXSTRING];
    int temp;
    int ptr;
    int16_t size;
    int label_x, label_y, label_orientation, label_size;
    
    ptr = iptr + 2; /* Initialize local pointer - skip byte count */
    
    memcpy(&byte_count, &input_buffer[iptr], 2);
    iptr = iptr + byte_count + 2;
    infileptr = infileptr + byte_count + 3;
    total_labels++;
    
    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nLabel x: %d", temp);
    if (opt_list) output_string(stemp);
    label_x = temp * SCALE + OFF_X;
    ptr += 2;

    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nLabel y: %d", temp);
    if (opt_list) output_string(stemp);
    label_y = temp * SCALE + OFF_Y;
    if (opt_geda) label_y = geda_y - label_y;
    ptr += 2;
    
    if (version == 2) {
        size=0; memcpy(&size, &input_buffer[ptr], 2);
        ptr += 2;
    } else {
        /* older version has a one-byte size */
        size=0; memcpy(&size, &input_buffer[ptr], 1);
        ptr += 1;
    }
    
    sprintf(stemp, "\nLabel orientation: ");
    if (opt_list) output_string(stemp);
    if (size > 0) {
        sprintf(stemp, "horizontal");
        label_orientation = 0;
    } else { 
        sprintf(stemp, "vertical");
        label_orientation = 3; /* NOTE: 'down' */
    }
    
    if (opt_list) output_string(stemp);
    
    temp = size; /* Convert uint16_t to int */
    temp = abs(temp);
    
    if (opt_list) { 
        sprintf(stemp, "\nLabel size: %d", temp);
        output_string(stemp);
    }
    label_size = temp;
    
    if (opt_list) { 
        sprintf(stemp, "\nLabel text: ");
        output_string(stemp);
    }
    
    temp = 0; memcpy(&temp, &input_buffer[ptr], 1);
    ptr++;
    memcpy(&stemp, &input_buffer[ptr], temp);
    stemp[temp] = '\0';
    if (opt_list) output_string(stemp);
    strncpy(label, stemp, MAXSTRING);
    ptr = ptr + temp;
    
    if (opt_kicad) {
        sprintf(stemp, "\nText Label %d %d %d %d ~ \n%s", label_x, label_y, label_orientation, (version == 2)?label_size * 60:label_size * 20, label);
        output_string(stemp);
    }  else if (opt_geda) {
        /* zero length net to accommodate the label */
        sprintf(stemp, "\nN %d %d %d %d 4", label_x, label_y, label_x, label_y);
        output_string(stemp);
        sprintf(stemp, "\n{\nT %d %d 5 %d 1 1 0 0 1\nnetname=%s\n}",label_x, label_y, label_size * 4, label); 
        output_string(stemp);
    }
    if (opt_list) { 
        sprintf(stemp, "\n");
        output_string(stemp);
    }
}

/* process_entry (Tag 08)
 * 
 */
void process_entry(void)
{
    int byte_count = 0;

    unsigned char stemp[MAXSTRING];
    int temp;
    int ptr;
    int entry_start_x, entry_start_y;
    int entry_end_offset_x, entry_end_offset_y; /* no end point provided by Orcad SDT - a fixed offset will be used */
    
    ptr = iptr + 2; /* Initialize local pointer - skip byte count */
    
    memcpy(&byte_count, &input_buffer[iptr], 2);
    iptr = iptr + byte_count + 2;
    infileptr = infileptr + byte_count + 3;
    total_entries++;
    
    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nEntry x: %d", temp);
    if (opt_list) output_string(stemp);
    entry_start_x = temp * SCALE + OFF_X;
    ptr += 2;

    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nEntry y: %d", temp);
    if (opt_list) output_string(stemp);
    entry_start_y = temp * SCALE + OFF_Y;
    if (opt_geda) entry_start_y = geda_y - entry_start_y;
    ptr += 2;
    
    temp = 0;  memcpy(&temp, &input_buffer[ptr], 1);
    sprintf(stemp, "\nEntry type: ");
    if (opt_list) output_string(stemp);
    if (temp & 0x1) {
        sprintf(stemp, "\\");
        entry_end_offset_x = -100;
        entry_end_offset_y = -100;
    } else {
        sprintf(stemp, "/");
        entry_end_offset_x = -100;
        entry_end_offset_y = 100;
    }
    if (opt_list) output_string(stemp);
    
    if (opt_kicad) {
        sprintf(stemp, "\nEntry Wire Line\n    %d %d %d %d", \
            entry_start_x, entry_start_y, entry_start_x + entry_end_offset_x, entry_start_y + entry_end_offset_y);
        output_string(stemp);
    } else if (opt_geda) {
        /* geda: bus entries are presented as nets */
        sprintf(stemp, "\nN %d %d %d %d 4", entry_start_x, entry_start_y, entry_start_x + entry_end_offset_x, entry_start_y - entry_end_offset_y);
        output_string(stemp);
    }
    if (opt_list) { 
        sprintf(stemp, "\n");
        output_string(stemp);
    }
}

/* process_dashed_line (Tag 09)
 * 
 */
void process_dashed_line(void)
{
    int byte_count = 0;

    unsigned char stemp[MAXSTRING];
    int temp;
    int ptr;
    int dashed_start_x, dashed_start_y, dashed_end_x, dashed_end_y;
    
    ptr = iptr + 2; /* Initialize local pointer - skip byte count */
    
    memcpy(&byte_count, &input_buffer[iptr], 2);
    iptr = iptr + byte_count + 2;
    infileptr = infileptr + byte_count + 3;
    total_dashed_lines++;
    
    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nDashed line start x: %d", temp);
    if (opt_list) output_string(stemp);
    dashed_start_x = temp * SCALE + OFF_X;
    ptr += 2;

    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nDashed line start y: %d", temp);
    if (opt_list) output_string(stemp);
    dashed_start_y = temp * SCALE + OFF_Y;
    if (opt_geda) dashed_start_y = geda_y - dashed_start_y; 
    ptr += 2;

    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nDashed line end x: %d", temp);
    if (opt_list) output_string(stemp);
    dashed_end_x = temp * SCALE + OFF_X;
    ptr += 2;

    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nDashed line end y: %d", temp);
    if (opt_list) output_string(stemp);
    dashed_end_y = temp * SCALE + OFF_Y;
    if (opt_geda) dashed_end_y = geda_y - dashed_end_y;
    ptr += 2;
    
    if (opt_kicad) {
        sprintf(stemp, "\nWire Notes Line\n    %d %d %d %d", \
            dashed_start_x, dashed_start_y, dashed_end_x, dashed_end_y);
        output_string(stemp);
    } else if (opt_geda) {
        /* L x1 y1 x2 y2 color width capstyle dashstyle dashlength dashspace */
        sprintf(stemp, "\nL %d %d %d %d 3 5 1 2 40 40", dashed_start_x, dashed_start_y, dashed_end_x, dashed_end_y);
        output_string(stemp);
    }
    if (opt_list) { 
        sprintf(stemp, "\n");
        output_string(stemp);
    }
}

/* process_power (Tag 0A)
 * 
 * kicad: power virtual components are individually defined and one
 * cannot change their names. components are similar to the following:
 *   Arrow: +BATT
 *   Line:  -BATT
 *   Circle: VDD
 *   Wave:   VAC
 * kicad: in order to compensate, a special library called pwr.lib
 * has been created, matching the corresponding orcad components.
 * 
 * geda: embedded components are created
 * 
 */
void process_power(void)
{
    int byte_count = 0;

    unsigned char stemp[MAXSTRING];
    unsigned char power_ref[MAXSTRING];
    unsigned char power_type[MAXSTRING];
    int temp;
    int ptr;
    int power_x, power_y;
    int a, b, c, d; /* orientation */
    int geda_angle; /* orientation */
    
    ptr = iptr + 2; /* Initialize local pointer - skip byte count */

    memcpy(&byte_count, &input_buffer[iptr], 2);
    iptr = iptr + byte_count + 2;
    infileptr = infileptr + byte_count + 3;
    total_power_entries++;
    
    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nPower x: %d", temp);
    if (opt_list) output_string(stemp);
    power_x = temp * SCALE + OFF_X;
    ptr += 2;

    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nPower y: %d", temp);
    if (opt_list) output_string(stemp);
    power_y = temp * SCALE + OFF_Y;
    if (opt_geda) power_y = geda_y - power_y;
    ptr += 2;
    
    temp = 0;  memcpy(&temp, &input_buffer[ptr], 1);
    sprintf(stemp, "\nPower type: ", temp);
    if (opt_list) output_string(stemp);
    ptr += 1;
    
    switch (temp & 0x3) {
        case 0:
            sprintf(stemp, "circle");
            strcpy(power_type, "pwr:circle");
            break;
        case 1:
            sprintf(stemp, "arrow");
            strcpy(power_type, "pwr:arrow");
            break;
        case 2:
            sprintf(stemp, "bar");
            strcpy(power_type, "pwr:bar");
            break;
        case 3:
            sprintf(stemp, "wave");
            strcpy(power_type, "pwr:wave");
            break;           
    }
    if (opt_list) output_string(stemp);
    
    sprintf(stemp, "\nPower orientation: ", temp);
    if (opt_list) output_string(stemp);
    
    switch (temp & 0xc) {
        case 0:
            sprintf(stemp, "top");
            a=1; b=0; c=0; d=-1;
            geda_angle = 0;
            break;
        case 4:
            sprintf(stemp, "bottom");
            a=-1; b=0; c=0; d=1;
            geda_angle = 180;
            break;
        case 8:
            sprintf(stemp, "left");
            a=0; b=-1; c=-1; d=0;
            geda_angle = 90;
            break;
        case 0xc:
            sprintf(stemp, "right");
            a=0; b=1; c=1; d=0;
            geda_angle = 270;
            break;           
    }
    if (opt_list) output_string(stemp);

    sprintf(stemp, "\nPower text: ");
    if (opt_list) output_string(stemp);
    
    temp = 0; memcpy(&temp, &input_buffer[ptr], 1);
    ptr++;
    memcpy(&stemp, &input_buffer[ptr], temp);
    stemp[temp] = '\0';
    if (opt_list) output_string(stemp);
    strncpy(power_ref, stemp, MAXSTRING);
    ptr = ptr + temp;

    if (opt_kicad) {
        subspaces(power_ref); /* kicad: spaces not allowed */
        sprintf(stemp, "\n$Comp\nL %s %s", power_type, power_ref);
        output_string(stemp);
        sprintf(stemp, "\nU 1 1 00000000");
        output_string(stemp);
        sprintf(stemp, "\nP %d %d", power_x, power_y);
        output_string(stemp);
        /* FIXME the positive offset for a PWR component label is not understood ! */
        sprintf(stemp, "\nF 0 \"#PWR?\" H %d %d 60 0001 C CNN", power_x, power_y + 200); /* invisible */
        output_string(stemp);
        sprintf(stemp, "\nF 1 \"%s\" H %d %d 60 0000 C CNN", power_ref, power_x, power_y + 150);
        output_string(stemp);
        sprintf(stemp, "\nF 2 \"\" H %d %d 60 0001 C CNN", power_x, power_y);
        output_string(stemp);
        sprintf(stemp, "\nF 3 \"\" H %d %d 60 0001 C CNN", power_x, power_y);
        output_string(stemp);
        sprintf(stemp, "\n    1    %d %d", power_x, power_y);
        output_string(stemp);
        sprintf(stemp, "\n    %d %d %d %d", a, b, c, d);
        output_string(stemp);
        sprintf(stemp, "\n$EndComp");
        output_string(stemp);
    } else if (opt_geda) {
        subspaces(power_ref); /* geda: spaces not allowed */
        /* C x y selectable angle mirror basename */
        subcolons(power_type);
        sprintf(stemp, "\nC %d %d 1 %d 0 %s-1.sym", power_x, power_y, geda_angle, power_type);
        output_string(stemp);
        if (strlen(power_ref)) {
            /* T x y color size visibility show_name_value angle alignment num_lines */
            sprintf(stemp, "\n{\nT %d %d 5 6 1 1 0 0 1\nrefdes=%s\n}", power_x - 100, power_y + 150, power_ref);
            output_string(stemp);
        }
    }
    if (opt_list) { 
        sprintf(stemp, "\n");
        output_string(stemp);
    }
}

/* process_text (Tag 0B)
 * 
 */
void process_text(void)
{
    int byte_count = 0;
    
    unsigned char stemp[MAXSTRING];
    unsigned char text[MAXSTRING];
    int temp;
    int ptr;
    int text_x, text_y, text_orientation, text_size; 
    int16_t size;
    
    ptr = iptr + 2; /* Initialize local pointer - skip byte count */
    
    memcpy(&byte_count, &input_buffer[iptr], 2);
    iptr = iptr + byte_count + 2;
    infileptr = infileptr + byte_count + 3;
    total_text_entries++;

    temp = 0;  memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nText Entry x: %d", temp);
    if (opt_list) output_string(stemp);
    text_x = temp * SCALE + OFF_X;
    ptr += 2;
    
    temp = 0;  memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nText Entry y: %d", temp);
    if (opt_list) output_string(stemp);
    text_y = temp * SCALE + OFF_Y;
    if (opt_geda) text_y = geda_y - text_y;
    ptr += 2;
    
    if (version == 2) {
        size=0; memcpy(&size, &input_buffer[ptr], 2);
        ptr += 2;
    } else {
        /* older version has a one-byte size FIXME: THIS IS NOT VERIFIED */
        size=0; memcpy(&size, &input_buffer[ptr], 1);
        ptr += 1;
    }
    
    sprintf(stemp, "\nText Entry orientation: ");
    if (opt_list) output_string(stemp);
    if (size > 0) {
        sprintf(stemp, "horizontal");
        text_orientation = 0; /* left justified */
    } else { 
        sprintf(stemp, "vertical");
        text_orientation = 3; /* 'down' - right justified */
    }
   if (opt_list) output_string(stemp);
    
    temp = size; /* Convert uint16_t to int */
    temp = abs(temp);
    
    sprintf(stemp, "\nText Entry size: %d", temp);
    if (opt_list) output_string(stemp);
    text_size = temp;
      
    sprintf(stemp, "\nText Entry text: ");
    if (opt_list) output_string(stemp);    
    temp = 0; memcpy(&temp, &input_buffer[ptr], 1);
    ptr++;
    memcpy(&stemp, &input_buffer[ptr], temp);
    stemp[temp] = '\0';
    if (opt_list) output_string(stemp);
    strncpy(text, stemp, MAXSTRING);
    ptr = ptr + temp;
    
    if (opt_kicad) {
        sprintf(stemp, "\nText Notes %d %d %d %d ~\n%s", \
            text_x, text_y, text_orientation, (version == 2)?text_size * 60:text_size * 20, text);
        output_string(stemp);
    } else if (opt_geda) {
        /* T x y color size visibility show_name_value angle alignment num_lines */
        sprintf(stemp, "\nT %d %d 9 %d 1 0 0 0 1\n%s", text_x, text_y, text_size * 4, text);
        output_string(stemp);
    }
    if (opt_list) { 
        sprintf(stemp, "\n");
        output_string(stemp);
    };
}

/* process_marker (Tag 0C)
 */
void process_marker(void)
{
    int byte_count = 0;

    unsigned char stemp[MAXSTRING];
    int temp;
    int ptr;
    int marker_x, marker_y;
    int nc_marker, stim_marker, vec_marker, trace_marker;
    int layout_marker, error_marker;
    unsigned char text[MAXSTRING];
    unsigned char marker_name[MAXSTRING];
    
    ptr = iptr + 2; /* Initialize local pointer - skip byte count */
    
    nc_marker = 0; stim_marker = 0; vec_marker = 0; trace_marker = 0; 
    layout_marker = 0; error_marker = 0;
    
    memcpy(&byte_count, &input_buffer[iptr], 2);
    iptr = iptr + byte_count + 2;
    infileptr = infileptr + byte_count + 3;
    total_markers++;
    
    temp = 0;  memcpy(&temp, &input_buffer[ptr], 1);
    sprintf(stemp, "\nMarker type: %x ", temp);
    if (opt_list) output_string(stemp);
    ptr += 1;
    strcpy(marker_name,"");
    switch (temp & 0x7) {
        case 0:
            sprintf(stemp, "Trace name");
            trace_marker = 1;
            strcpy(marker_name, "trace");
            break;
        case 1:
            sprintf(stemp, "Vector column");pwr:
            vec_marker = 1;
            strcpy(marker_name, "vector");
            break;
        case 2:
            sprintf(stemp, "Stimulus");
            stim_marker = 1;
            strcpy(marker_name, "stimulus");
            break;
        case 3:
            sprintf(stemp, "Error");
            error_marker = 1;
            break;          
        case 4:
            sprintf(stemp, "No Connect");
            nc_marker = 1;
            break;
        case 5: 
            sprintf(stemp, "Layout directive");
            layout_marker = 1;
            strcpy(marker_name, "layout");
            break;
        default:
            sprintf(stemp, "UNKNOWN");
    }
    if (opt_list) output_string(stemp);
    
    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nMarker x: %d", temp);
    if (opt_list) output_string(stemp);
    marker_x = temp * SCALE + OFF_X;
    ptr += 2;

    temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
    sprintf(stemp, "\nMarker y: %d", temp);
    if (opt_list) output_string(stemp);
    marker_y = temp * SCALE + OFF_Y;
    if (opt_geda) marker_y = geda_y - marker_y;
    ptr += 2;
    
    if (trace_marker || stim_marker || vec_marker) {
        temp = 0; memcpy(&temp, &input_buffer[ptr], 2);
        sprintf(stemp, "\nData: %4.4x ", temp);
        if (opt_list) output_string(stemp);
        ptr += 2;
    } else {
        sprintf(stemp, "\n");
        if (opt_list) output_string(stemp);
    }
    if (trace_marker || stim_marker || layout_marker || error_marker || vec_marker) {
        temp = 0; memcpy(&temp, &input_buffer[ptr], 1);
        ptr++;
        memcpy(&stemp, &input_buffer[ptr], temp);
        stemp[temp] = '\0';
        if (opt_list) output_string(stemp);
        strncpy(text, stemp, MAXSTRING);
        ptr = ptr + temp;
    }
    
    if (opt_kicad) {
        if (nc_marker) {
            sprintf(stemp, "\nNoConn ~ %d %d", marker_x, marker_y);
            output_string(stemp);
        }
        if (trace_marker || stim_marker || layout_marker || vec_marker) {
            sprintf(stemp, "\n$Comp\nL pwr:%s ~", marker_name);
            output_string(stemp);
            sprintf(stemp, "\nU 1 1 00000000");
            output_string(stemp);
            sprintf(stemp, "\nP %d %d", marker_x - 100, marker_y);
            output_string(stemp);
            sprintf(stemp, "\nF 0 \"~\" H %d %d 60 0001 C CNN", marker_x, marker_y + 200); /* invisible */
            output_string(stemp);
            sprintf(stemp, "\nF 1 \"%s\" H %d %d 60 0000 C CNN", text, marker_x, marker_y + 150);
            output_string(stemp);
            sprintf(stemp, "\nF 2 \"\" H %d %d 60 0001 C CNN", marker_x, marker_y);
            output_string(stemp);
            sprintf(stemp, "\nF 3 \"\" H %d %d 60 0001 C CNN", marker_x, marker_y);
            output_string(stemp);
            sprintf(stemp, "\n    1    %d %d", marker_x - 100, marker_y);
            output_string(stemp);
            sprintf(stemp, "\n    1 0 0 -1"); /* no rotation */
            output_string(stemp);
            sprintf(stemp, "\n$EndComp");
            output_string(stemp);
        }
        if (error_marker) {
            sprintf(stemp, "\nText Notes %d %d 0 80 ~\n%s%s", \
                marker_x, marker_y - 20, error_marker?"?? ":"", text);
            output_string(stemp);
        }
    } else if (opt_geda) {
        /* NOTE: No Connect marker is not handled at present */
        if (trace_marker || stim_marker || layout_marker || vec_marker) {
            /* C x y selectable angle mirror basename */
            sprintf(stemp, "\nC %d %d 1 0 0 pwr_%s-1.sym", marker_x, marker_y, marker_name);
            output_string(stemp);
            /* T x y color size visibility show_name_value angle alignment num_lines */
            sprintf(stemp, "\n{\nT %d %d 5 6 1 1 0 0 1\nrefdes=%s\n}", marker_x - 100, marker_y + 150, text);
            output_string(stemp);
        }
    }
    if (opt_list) { 
        sprintf(stemp, "\n");
        output_string(stemp);
    }
}

/* process_end_of_file (Tag 0F)
 * byte count is zero
 */
void process_end_of_file(int max_iptr)
{
    unsigned char stemp[MAXSTRING];
    int byte_count = 0;
    int i;
    memcpy(&byte_count, &input_buffer[iptr], 2);
    iptr = iptr + byte_count + 2;
    infileptr = infileptr + byte_count + 3;
    if (opt_list)  {
        sprintf(stemp, "\norcad_convert: EOF: %x %lld\n", infileptr, infileptr);
    } else if (opt_kicad) {
        sprintf(stemp, "\n$EndSCHEMATC");
    } else if (opt_geda) {
        stemp[0] = '\0';
    }
    output_string(stemp);
    
    /* At this point, iptr points to Component names
    * preceded by a single-byte count.
    */
    sprintf(stemp, "\norcad_convert: Component names\n");
    if (opt_list) output_string(stemp);
    i=0;
    while (iptr < max_iptr) {
        byte_count = 0;
        memcpy(&byte_count, &input_buffer[iptr++], 1);
        memcpy(&stemp, &input_buffer[iptr], byte_count);
        stemp[byte_count] = 0x0A;
        stemp[byte_count + 1] = '\0';
        if (opt_list) output_string(stemp);
        iptr = iptr + byte_count;
        infileptr = infileptr + byte_count +1;
        i++;
    }

    sprintf(stemp, "\norcad_convert: Number of component names: %d\n", i);
    if (opt_list) output_string(stemp);

}

void output_statistics(char *input_f)
{
    total_items = total_sheets + total_components + total_wires + total_busses + total_junctions + total_module_ports +\
        total_labels + total_entries + total_dashed_lines + total_power_entries + total_text_entries + total_markers;

    printf("\norcad_convert: Contents of file: %s\n", input_f);
    printf("\n");
    printf("\norcad_convert: Sheets:        %d\n", total_sheets);
    printf("\norcad_convert: Components:    %d\n", total_components);
    printf("\norcad_convert: Wires:         %d\n", total_wires);
    printf("\norcad_convert: Busses:        %d\n", total_busses);    
    printf("\norcad_convert: Junctions:     %d\n", total_junctions);
    printf("\norcad_convert: Module ports:  %d\n", total_module_ports);
    printf("\norcad_convert: Labels:        %d\n", total_labels);
    printf("\norcad_convert: Entries:       %d\n", total_entries);    
    printf("\norcad_convert: Dashed lines:  %d\n", total_dashed_lines);
    printf("\norcad_convert: Power entries: %d\n", total_power_entries);
    printf("\norcad_convert: Text entries:  %d\n", total_text_entries);
    printf("\norcad_convert: Markers:       %d\n", total_markers);
    printf("\n");
    printf("\norcad_convert: Total items:   %d\n", total_items);
    printf("\n");
}
