/* orcad_lib_convert: Lists and translates decomposed (.src) OrCAD SDT library files
 *
 * Copyright (c) 2020 Dimitri Marinakis
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License Version 2
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Usage:
 *	 orcad_lib_convert [options] [.src file] [output_file]
 *
 * Options:
 *	-h -- display program information
 *  -g -- geda output
 *  -k -- kicad output
 *  -p -- power pins are shown (geda)
 *	-s -- file statistics
 *	-V -- processing debug information is printed as in-line comments
 *
 * Returns: exit code 1 if unsuccesful
 *
 * Make: gcc orcad_lib_convert.c -lm -o orcad_lib_convert
 * 
 * Examples:
 * 
 *  ./orcad_lib_convert -k TTL.SRC ttl.lib (kicad library)
 *
 *  ./orcad_lib_convert -g TTL.SRC (geda .sym files created under directory ./TTL)
 * 
 *  Also produces TTL.SRC.txt with component names and x/y sizes for use by the orcad_convert program.
 *
 * References:
 *   OrCAD SDT manual
 *   KiCad library format
 *   gEDA file format
 *   gEDA master attribute document
 * 
 * History: 
 *   01-Mar-2020 dm
 *   25-Mar-2020 dm -- set anchor at the left corner of a component instead of the bottom left corner
 *   04-Apr-2020 dm -- generates component width/height file (.txt suffix) for use in rotation calculations
 *   25-Apr-2020 dm -- adds library name to component width/height file
 *   31-Apr-2020 dm -- geda output
 */

#include "orcad_lib_convert.h"

/****************** PROTOTYPES FROM THE TOP DOWN *******************/

int orcad_lib_convert (char * input_file, char * output_file);
int process_input (int input_count);
off_t fsize(const char *filename);
double angle(int cx, int cy, int ex, int ey);
void write_component(char *cname, int xs, int ys);
void write_part_data(char *data);
int output_string (char * bstring);
void subspaces(char * string);
void noticks(char * string);
int existing_vector(char * part);
void write_slot_data(int p);
void process_line(char * input_line);
void output_statistics(char *input_f);

/****************** PROGRAM VERSION INFORMATION  *******************/
char *prog_version="1.9";

/****************** COMMAND LINE OPTIONS  **************************/
int opt_verbose = 0;
int opt_list = 1; /* default option */
int opt_statistics = 0;
int opt_kicad = 0;
int opt_geda = 0;
int opt_power = 0;


/****************** GLOBAL VARIABLES *******************************/

unsigned char prog_name[MAXSTRING];

/* off_t - print with %lld 
 */
off_t iptr = 0;	/* input buffer pointer, reset before every read */
off_t optr = 0;	/* output buffer pointer, reset after every write */
off_t infileptr = 0; /* file input pointer, updated with last input */
off_t infilesize = 0; /* input file size */

/* For saving and reproducing vector information of components */
unsigned char existing_part[MAXPARTS][MAXSTRING];
off_t vector_ptr[MAXPARTS];  /* start of normal representation VECTOR data */
off_t convert_ptr[MAXPARTS];  /* start of equivalent representation VECTOR data */
off_t iptr_save = 0;   /* variable to store the current input buffer location */
int part_in_table = 0;
int convert_state =0;
int increment_part_in_table = 0;
int based_on_previous = 0;

unsigned char input_buffer[INRECSIZE];
unsigned char output_buffer[OUTRECSIZE];

int processing = 0; /* processing flag may be modified by individual processing routines */
int processing_state = START;

int components = 0;  /* number of components in library */

unsigned char prefixes[MAXPREFIXES][MAXSTRING]; /* library prefixes */
int prefixidx;

unsigned char partnames[MAXPARTNAMES][MAXSTRING]; /* part name and aliases */
int partnameidx;
int gridarray;
int dmge; /* kicad: De Morgan equivalent graphic shape (1 or 2), geda: used to generate "-2.sym" file suffix*/
int x_size = 0, y_size = 0;   /* part body dimensions */
int parts_per_package;

unsigned char stemp[MAXSTRING];
unsigned char flink[MAXSTRING]; /* geda: used to created symlinks to component files */
int enddraw_pending = 0;    /* kicad: DRAW started but ENDDRAW has not been output yet */

int pin_sequence = 0; /* geda: sequential numbering as pins are encountered */

int slot_data[MAXPARTSPERPACKAGE][MAXPINSPERPART]; /* geda: stores the pins for each part */
int gslots = 0;   /* geda (slot_data): slot index */
int gpins = 0;  /* geda (slot_data): pin index */


#define NEWPART(pindex) \
{ \
    /* Commands to be used before every PARTNAME state change \
     * index should be either 0 (part name identifier) or 1 (first alias) \
     * demorgan equivalent (dmge) should be either 1 (default) \
     * or 2 (convert) \
     */ \
    partnameidx = pindex; \
    parts_per_package = 0; \
    gridarray = 0; \
    dmge = 1; \
    convert_state = 0; \
    based_on_previous = 0; \
    if (increment_part_in_table) {part_in_table++; increment_part_in_table = 0;} \
    strcpy(ref,"U"); \
}

FILE *comp; /* component size file */
unsigned char lib_name[MAXSTRING];  /* holds the library file name, no extension */

FILE *partfile = NULL; /* part file for geda */

/****************** HANDLE ARGUMENTS AND SYNTAX *******************/

void usage_info(void)
{
	printf("\n%s: Component library (.src) converter ver. %s\n", prog_name, prog_version);
	printf("\n%s: Copyright (c) 2020 Dimitri Marinakis", prog_name);
    printf("\n%s: This program is free software; you are encouraged to redistribute it under", prog_name);
	printf("\n%s: the terms of the GNU General Public License (version 2).\n", prog_name);
	printf("\n%s: This program comes with ABSOLUTELY NO WARRANTY. See the GNU General Public", prog_name);
	printf("\n%s: License in the file named `COPYING' for more details.\n", prog_name);
	printf("\n%s: Syntax: orcad_lib_convert [options] [.src file] [output_file]\n", prog_name);
	printf("\n%s:         (if .src_file is not given, program reads from standard input)\n", prog_name);
	printf("\n%s: -h -- display program information", prog_name);
    printf("\n%s: -k -- kicad output", prog_name);
    printf("\n%s: -g -- geda output", prog_name);
    printf("\n%s: -p -- power pins are shown (geda only)", prog_name);
 	printf("\n%s: -s -- statistics", prog_name);
	printf("\n%s: -V -- additional processing information as in-line comments", prog_name);
	printf("\n\n");
}

int main(int argc, char * argv[])
{
	int idx, i;

    strncpy(prog_name, argv[0], MAXSTRING - 1);

	while(argc > 1 && argv[1][0] == '-') {
		switch(argv[1][1]) {
			case 'V':
				opt_verbose = 1;
				break;
			case 's':
				opt_statistics = 1;
				break;
            case 'g':
                opt_geda = 1;
                opt_list = 0;   /* clear default list option */
                break;
            case 'p':
                opt_power = 1;  /* show power pins - geda */
                break;
            case 'k':
                opt_kicad = 1;
                opt_geda = 0;   /* kicad overrides geda */
                opt_list = 0;   /* clear default list option */
                break;
			case '?':
			case 'h':
                /* fall through for unknown option */
			default:
				usage_info();
				exit(0);
		}
	argc-- ;
	argv++ ;
	}

	switch(argc) {
		case 1:
			return orcad_lib_convert("", "");
			break;
		case 2:
			return orcad_lib_convert(argv[1], "");
			break;
		default:
			return orcad_lib_convert(argv[1], argv[2]);
 	}

}


/****************** FILE READ-WRITE LOOP **************************/

int orcad_lib_convert(char *input_file, char *output_file)
{
	int i, j;  /* counters */
	int infile;    /* input file descriptor */
	int outfile;   /* output file descriptor */

	off_t icnt;   /* input file bytes read in input buffer */
	off_t ocnt;   /* output file bytes written */
	off_t bytes_to_process;	/* bytes read minus bytes already processed */

	int wehaveinput;    /* flag for while */
	int process_return;    /* process return code */

	char new_file[MAXFILELEN];
    char comp_file[MAXFILELEN];
    
	struct stat filestat;  /* used to obtain the file size */
	
	/* Open input and output files
	 */
	if (strlen(input_file) == 0) {
		infile = STDIN_FILENO;
	} else {
		infile = open(input_file, O_RDONLY);
        infilesize = fsize(input_file);
	}
	if (infile < 0 || infilesize < 0) {
        printf("\n%s: input file open error", prog_name);
		return(1);
	} else {
		if (strlen(output_file) == 0) {
			outfile = STDOUT_FILENO;
		} else {
			if (strlen(output_file) < MAXFILELEN) {
				strcpy(new_file, output_file);
			} else {
				printf("\n%s: output filename too long - exiting\n", prog_name);
				return(1);
			}
			outfile = open(new_file, O_WRONLY | O_CREAT | O_TRUNC);
		}
		if (outfile < 0) {
			printf("\n%s: output file create error", prog_name);
			close(infile);
			return(1);
		}
	}
	
	/*
     * Open component information output file
     * and set the library name (same as input filename, no extension).
     * (if the file exists, it is overwritten)
     */
    int scanidx = strlen(input_file);
    int first_char = 0;
    int dot_char = 0;
    if (scanidx) {
        scanidx--;
        while (scanidx) {
            if (input_file[scanidx] == '.') {
                dot_char = scanidx;
                scanidx = 0;    /* exit loop */
            } else {
                scanidx--;
            }
        }
        if (dot_char) {
            /* dot found, copy string to previous slash, if any */
            scanidx = dot_char;
            while (scanidx) {
                if (input_file[scanidx] == '/') {                           
                    first_char = ++scanidx;
                    scanidx = 0;
                } else {
                    scanidx--;
                }
            }
            strncpy(lib_name, &input_file[first_char], dot_char - first_char);
            lib_name[dot_char - first_char + 1] = '\0';
        } else {
            lib_name[0] = '\0';
        }

        sprintf(comp_file, "./%s.txt", input_file);
        comp = fopen(comp_file, "w"); /*  file */
        if (comp == NULL) {
            printf("\n%s: component size output file create error", prog_name);
            return(1);
        }
        
        if (opt_geda) {
            if (mkdir(lib_name, 0777) == -1) {
                sprintf(stemp, "\n%s: %s", prog_name, lib_name);
                perror(stemp);
                /* ignore errors but display notification */
            }
        }
    }
    
	/* Initialize processing variables
	 */
    components = 0;
    
    /* Initialize BITMAP/CONVERT variables
     * which may be used by other parts
     */
    for (i = 0; i < MAXPARTS; i++) {
        existing_part[i][0] = '\0';
        vector_ptr[i] = 0;
        convert_ptr[i] = 0;
    }
    part_in_table = 0;
    convert_state = 0;
    increment_part_in_table = 0;
    
	/*  Read, process and write file blocks,
	 *  Change file mode (equivalent to chmod 644 output_file),
	 *  and return.
     * 
     *  NOTE: a single read is used for this line-oriented application
     *  NOTE: input files should fit in input_buffer (INRECSIZE)
	 */
	wehaveinput = 1;
	while (wehaveinput) {
		icnt = read(infile, input_buffer, sizeof(input_buffer));
		if (icnt == 0) wehaveinput = 0;

        /* Processing loop */
		iptr = 0;
		while ((icnt > 0) && (iptr < icnt)) {
    
			bytes_to_process = icnt - iptr;
			process_return = process_input(bytes_to_process);          

            /* Write processed data and reset output buffer pointer */
			if (optr >= 0) { /* processing could return with zero output bytes */
				ocnt = write(outfile, output_buffer, optr);
				optr = 0;
				if (ocnt < 0) {
                    perror("\n ");
					printf("\n%s: output file write error", prog_name);
					wehaveinput = 0;
				}
			} else if (process_return != -2) {
				if (opt_verbose) printf("\n%s: no more bytes to write", prog_name);
				wehaveinput = 0;    /* signal no more input */
			}
		}
	}
	
	/* Output statistics, if requested,
	 * close input and output files,
	 * make output file readable.
	 */
	close(infile);
	if (opt_statistics || opt_verbose) output_statistics(input_file);
	if (opt_verbose) printf("\n%s: processing complete %d\n", prog_name, enddraw_pending);

	if (outfile == STDOUT_FILENO) printf("\n");

	if (close(outfile)) {
		perror("\n ");
        printf("%s: output file close", prog_name);
		return(1);
	}
	if (outfile != STDOUT_FILENO) {
		if (chmod(new_file, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) {
			perror("\n ");
            printf("\n%s: output file chmod", prog_name);
			return(1);
		}
	}
	if (comp) {
        fclose (comp); /* close component output file */
    }
	return(0);
}

/****************** PROCESSING *************************************/

/* process_input:
 * Processes <input_count> bytes in <input_buffer>
 * Returns: bytes written, -1 for EOF
 * orcad: comment items enclosed in curly brackets {} are removed
 * orcad: curly brackets in literal strings (e.g. TEXT) are maintained
 * Changes: iptr, infileptr
 */
int process_input (int input_count)
{
	unsigned char inchar;
	unsigned int outcode;
	unsigned char current;
	int iptr_max;       /* holds the calculated maximum input pointer value */
	int return_code;    /* bytes written; error if negative */
	
	unsigned char line[MAXSTRING]; /* holds a single line with non-zero length */
    int lineptr;
    int ignore_comment;
    int literal;

	return_code = 0;

	processing = 1;
	iptr_max = iptr + input_count;
    
    lineptr = 0;
    line[0] = '\0';
    ignore_comment = 0;
    processing_state = START;
    partnameidx = 0; /* will be initialized again, on state change */

    while (processing) {
		if ((iptr < INRECSIZE) && (iptr < iptr_max)) {
			inchar = input_buffer[iptr++];

            /* If a carriage return is found process a non-zero length line */
            if (inchar == 0x0d && strlen(line)) {
                line[lineptr] = '\0';
                process_line(line);
                lineptr = 0;
            } else {
                switch (inchar) {
                    case 0x0a:  /* A line feed and a second carriage return (0xa, 0xd) may follow */
                    case 0x0d:
                        literal = 0; /* reset literal indicator */
                        ignore_comment = 0; /* reset ignore_comment indicator */
                        lineptr = 0;
                        line[0] = '\0';
                        break;
                    case 0x1a:  /* EOF */
                        if (opt_verbose) printf("\n\n%s: EOF (0x1A) found - iptr = %4.4x, iptr_max = %4.4x, infileptr = %4.4x, optr = %4.4x\n",\
                                prog_name, iptr, iptr_max, infileptr, optr);
                        /* 
                         * Handle leftovers 
                         */
                        if (opt_kicad && enddraw_pending) {
                            if (opt_verbose) {sprintf(stemp, "\n# Processing: exit"); output_string(stemp);}
                                if (processing_state == PINS) {
                                    /* have not gone through the BITMAP or VECTOR state,
                                    * must draw a rectangle.
                                    * kicad: S X1 Y1 X2 Y2 part dmg pen fill Rectangle, from X1,Y1 to X2,Y2.
                                    * kicad: 0,0 is top left corner of the component
                                    * kicad: x increases to the right, y increases upwards 
                                    * part = 0 (applies to all parts)
                                    */
                                    sprintf(stemp, "\nS 0 0 %d %d 0 %d 0 f", x_size, -y_size, dmge);
                                    output_string(stemp);
                                }
                            sprintf(stemp, "\nENDDRAW");
                            output_string(stemp);
                            sprintf(stemp, "\nENDDEF\n");
                            output_string(stemp);
                        } else if (opt_geda && enddraw_pending) {
                            if (partfile) {fclose(partfile); partfile = NULL;} 
                            //FIXME: geda: MAY NEEED RECTANGLE TO COMPLETE PART
                        }
                        if (opt_kicad) {
                            sprintf(stemp, "\n#End Library\n");
                            output_string(stemp);
                        }
                        processing = 0;
                        if (strlen(line)) return_code = -1; /* Error if EOF encountered without an end of line */
                        break;
                    case 0x27: /* single quote */
                        literal = literal?0:1; /* if literal string is open, close it, else open it */
                        line[lineptr++] = inchar;
                        break;
                    /* 
                     * Curly brackets delimit comments, unless found inside literal strings
                     */
                    case '{':
                        if (literal == 0) {
                            ignore_comment = 1;
                        } else {
                            line[lineptr++] = inchar;
                        }
                        break;
                    case '}':
                        if (literal == 0) {
                            ignore_comment = 0;
                        } else {
                            line[lineptr++] = inchar;
                        }
                        break;
                    default:
                       if (ignore_comment == 0) line[lineptr++] = inchar;
                }
            }
		} else {
			/* Finished processing all input */
            printf("\n%s: processing finished\n", prog_name);
			processing = 0;
		}
	} /* end while processing*/

	return return_code;
}/* iptr should be zero to begin with */



/****************** LIBRARY FUNCTIONS ******************************/

off_t fsize(const char *filename) {
    struct stat st;

    if (stat(filename, &st) == 0)
        return st.st_size;

    fprintf(stderr, "\n%s: Cannot determine size of %s: %s\n", prog_name, filename, strerror(errno));

    return -1;
}

double angle(int cx, int cy, int ex, int ey) 
{
    /* Calculates angle from center cx,cy to end ex,ey
     * returns normalized angle in tenths of a degree.
     */
    double dy = ey - cy;
    double dx = ex - cx;
    double theta = atan2(dy, dx); // range (-PI, PI]
    theta *= 180.0 / M_PI; // rads to degs, range (-180, 180]
    if (theta < 0) theta = 360 + theta; // range [0, 360)
    return rint(theta*10);
}

void write_component(char *cname, int xs, int ys)
{
    unsigned char comp_string[MAXSTRING];
    if (comp) {
        sprintf(comp_string, "%s %d %d %s\n", cname, xs, ys, lib_name);
        fputs(comp_string, comp);
    }
}

void write_part_data(char *data)
{
    unsigned char part_string[MAXSTRING];
    if (partfile) {
        sprintf(part_string, "%s", data);
        fputs(part_string, partfile);
    }
}

int output_string (char * bstring)
{
	/* Writes strlen(bstring) bytes to output_buffer
	 * kicad: modifies optr
     * geda: information is output to part file
     * NOTE: geda: should not be part of pin attributes
	 */
	int i;
	int len;
    int return_code;
    
    return_code = 0;
    
    if (opt_geda) {
        /* geda: information written to part file or standard output*/
        if (partfile) write_part_data(bstring);
        else printf(bstring);
    } else {
        /* kicad: write to output buffer (single write) */
        len = strlen(bstring);
        if (len && (optr < OUTRECSIZE)) {
            for (i=0; i<len; i++) output_buffer[optr++] = bstring[i];
        } else {
            return_code = -1;
        }
    }
    return return_code;
}

void subspaces(char * string)
{
   /* Substitutes spaces, hyphens ('-') and slashes ('/') with '_' for component names
    * kicad: spaces are not allowed in component names
    * geda: hyphens and slashes are not allowed in component names
    */
    for (int i=0; i<strlen(string); i++) {
        if(string[i] == 0x20 || string[i] == 0x2d || string[i] == 0x2f) {
            string[i] = '_';
        }
    }
}

void noticks(char * string)
{
    /* Replaces single quotes in input string
     * with spaces
    */
    for (int i=0; i<strlen(string); i++) {
        if(string[i] == '\'') {
            string[i] = 0x20;
        }
    }
}

int existing_vector(char * part)
{
    /* We are searching for one more part than what has been stored
     * in the table to compensate for the case where the same component is
     * using the BITMAP for CONVERT
     */
    int index = -1; /* negative value means not found */
    for (int i=0; i<part_in_table+1; i++) {
        if (strcmp(existing_part[i], part) == 0) {
            index = i;
            break;
        }
    }
    return index;
}

void write_slot_data(int p)
{
    /* geda: write slot definition data collected in the slot_data array
     */
    if (parts_per_package > 1) {
        int i, j;
            for (i=0; i<parts_per_package; i++) {
            /* slotdef: detached attribute color, hidden */
            sprintf(stemp, "\nT %d %d 8 3 0 0 0 0 1", 200,  150 + 50*i); /* place on top of package */
            write_part_data(stemp);
            sprintf(stemp, "\nslotdef=%d:%d",i+1,slot_data[i][0]); /* write first pin data */
            write_part_data(stemp);
            for (j=1; j<p; j++) {
                sprintf(stemp, ",%d", slot_data[i][j]); /* write remaining pins */
                write_part_data(stemp);
                }
        }
    }
}

/****************** PROCESSING FUNCTIONS ******************************/

void process_line(char *input_line)
{
   /* Breaks orcad SDT component source input line into tokens, 
    * removing comments delimited by curly brackets { }
    * and generates output depending on processing_state
    */
    int inptr;
    unsigned char tokens[MAXTOKENS][MAXSTRING];
    int tokenidx;
    char *token;
    int tokenlen;
    int i, j;
    unsigned char ref[MAXSTRING];
    int part_name_invisible;
    unsigned char pin_name[MAXSTRING]; /* pin name, '~name' marks inverted, a single '~' marks no name*/
    int pin_pos; /* relative pin position */
    int pin_number[MAXPARTSPERPACKAGE]; /* pin number for each part, indexed by parts_per_package */
    unsigned char pin_gridarray[MAXSTRING]; /* pin name for GRIDARRAY */
    unsigned char pin_orientation[2]; /* Up, Down, Left, or Right */
    unsigned char pin_type[2]; /* Input, Output, Bidirectional, Tristate, Passive, open Collector, open Emitter, Not connected, Unspecified, W - power input, w - power output */
    unsigned char pin_shape[MAXSTRING]; /* kicad: if absent, line, or Inverted, Clock, Low input, V - output low; invisible if first letter is N */
    int pin_length; /* length of pin line */

              
    tokenidx = 0;
    prefixidx = 0;
    
    if (opt_list) {
        sprintf(stemp, "\n");
        output_string(stemp);
    }
    token = strtok(input_line, " "); 
    if (token != NULL) strcpy(tokens[tokenidx++], token);
    
    while (token != NULL) { 
        sprintf(stemp, "%s ", token);
        if (opt_list) output_string(stemp);
        token = strtok(NULL, " ");
        if (token != NULL) strcpy(tokens[tokenidx++], token);
    }
    
    /* At this point, the tokens array holds the broken down input line and
     * tokenidx holds the number of tokens on the line.
     * 
     * The orcad SDT component source file structure is:
     * PREFIX
     * COMPONENTS
     *  COMPONENT NAMES AND ALIASES
     *  COMPONENT SIZE AND PARTS PER PACKAGE
     *  PIN DEFINITIONS
     *  OPTIONAL BITMAP
     *  OPTIONAL VECTOR DEFINITION
     *  OPTIONAL CONVERT (DE MORGAN EQUIVALENT)
     *  OPTIONAL CONVERT PIN DEFINITIONS
     *  OPTIONAL CONVERT BITMAP
     *  OPTIONAL CONVERT VECTOR DEFINITION
     * EOF (^Z - 0x1A)
     */
    switch (processing_state) {
        case START:
            /* orcad: PREFIX entry is mandatory
             */
            if ((tokenidx == 1) && (strcmp(tokens[0], "PREFIX") == 0)) {
                processing_state = PREFIX;
                if (opt_kicad) {
                    sprintf(stemp, "EESchema-LIBRARY Version 2.4\n#encoding utf-8");
                    output_string(stemp);
                }
                
            }
            break;

        case PREFIX:
            /* orcad: up to 16 prefixes may be defined
             * kicad: PREFIX information is not used 
             */
            if (tokenidx == 0) {
                /* do nothing */
            } else if ((tokenidx == 1) && strcmp(tokens[0], "END") == 0) { 
                NEWPART(0)
                processing_state = PARTNAME;
            } else {
                /* combine tokens depending on output format 
                 */
                sprintf(prefixes[prefixidx], "\n");
                for (i=0; i<tokenidx; i++) {
                    strcat(prefixes[prefixidx], tokens[i]);
                    strcat(prefixes[prefixidx], " ");
                }
                /* NOTE: for information only -- 
                 * kicad: prefix is not used 
                 */
                if (opt_list) output_string(prefixes[prefixidx]); 
                prefixidx++;
                if (prefixidx > MAXPREFIXES) {
                    if (opt_list) {
                        sprintf(stemp, "\n%s: ERROR -- Maximum number of prefixes exceeded!", prog_name);
                        output_string(stemp);
                    }
                    /* Error - try to continue, anyway */
                    NEWPART(0);
                    processing_state = PARTNAME;
                }
            }
            break;

        case PARTNAME:
            /* 
             * Collect part name, reference and symbol size information
             */
            if (tokenidx == 0) {
                /* 
                 * Ignore empty lines, stay in the PARTNAME state 
                 */
            } else if (strstr(tokens[0], "\'")) {
                /* orcad: a single quote in the PARTNAME state
                 * marks a partname detected
                 * combine all space-separated tokens to handle
                 * partnames with spaces inside; 
                 * NOTE: spaces and hyphens are replaced by underscores (subspaces)
                 */
                int spidx=1;
                while (tokenidx > spidx) {
                        strcat(tokens[0], "_");
                        strcat(tokens[0], tokens[spidx++]);
                }
                tokenlen = strlen(tokens[0]) - 2;
                strncpy(partnames[partnameidx], &tokens[0][1], tokenlen);
                partnames[partnameidx][tokenlen] = '\0';
                subspaces(partnames[partnameidx]);
                if (opt_verbose) {
                    sprintf(stemp, "\n# PARTNAME (part) %s", tokens[0]);
                    output_string(stemp);
                }
                partnameidx++;
                if (partnameidx >= MAXPARTNAMES) fprintf(stderr,"\nERROR: Maximum number of partnames exceeded!");
            } else if (tokenidx == 2) {
                /* 
                 * REFERENCE is the only case with two tokens 
                 * (if not explicitly listed, 'U' is used)
                 */
                if (opt_verbose) {
                    sprintf(stemp, "\n# PARTNAME (ref) %s %s", tokens[0], tokens[1]);
                    output_string(stemp);
                }
                if (strcmp(tokens[0], "REFERENCE") == 0) {
                    tokenlen = strlen(tokens[1]) - 2;
                    strncpy(ref, &tokens[1][1], tokenlen);
                    /* FIXME: remove question mark, if present */
                }
            } else if (tokenidx == 3) {
                /* 
                 * Package information has three tokens: 
                 * x_size, y_size, parts_per_package or GRIDARRAY 
                 */
                if (opt_verbose) {
                        sprintf(stemp, "\n# PARTNAME (package) %s %s %s", tokens[0], tokens[1], tokens[2]);
                        output_string(stemp);
                }
                x_size = atoi(tokens[0]) * SCALE;
                y_size = atoi(tokens[1]) * SCALE;
                parts_per_package = atoi(tokens[2]);
                part_name_invisible = 0;
                if (parts_per_package == 0) { 
                    //FIXME: CHECK IF parts_per_package = 1 is necessary
                    part_name_invisible = 1;
                }
                if (strcmp(tokens[2], "GRIDARRAY") == 0) {
                    gridarray = 1;
                    part_name_invisible = 0;
                    parts_per_package = 1;
                }
                if (strcmp(tokens[2], "IEEE") == 0) {
                    part_name_invisible = 0;
                    parts_per_package = 1;
                }
                
                processing_state = PARTNAME_COMPLETE; /* PARTNAME -> PARTNAME_COMPLETE */
            }
            break;

            case PARTNAME_COMPLETE:
                /*
                 * Output collected partname information and prepare for PINS
                 */
                if (opt_kicad || opt_geda) {
                    /* kicad: part is defined with the anchor (0, 0) at the top left corner 
                     * kicad: x increases to the right, y increases upwards relative to anchor
                     * geda: same as kicad, but part needs to be translated 
                     * geda: in the geda symbol editor for it to be fully visible
                     * 
                     * Reference and part number are placed at the top and bottom of the part, 
                     * respectively. If the part is 200 mils wide or less, this information 
                     * is placed to the right of the part. 
                     * Horizontal and Vertical alignment is Left, Bottom
                     */
                    int ref_pos_x, ref_pos_y;
                    int part_value_pos_x, part_value_pos_y;

                                     
                    if (opt_kicad) {
                        /* kicad: DEF [~]name ref 0 offs Y Y 1 F N -- fixed pin name offset 50 mils*/
                        sprintf(stemp, "\nDEF %s%s %s 0 50 Y Y %d F N", \
                            (part_name_invisible?"~":""), partnames[0], ref, parts_per_package);
                        output_string(stemp);
                    }
                    if (x_size <= 200) {
                        ref_pos_x = x_size + 40;
                        ref_pos_y = -(y_size / 2) + 10;  /* Bottom alignment */
                        part_value_pos_x = x_size + 40;
                        part_value_pos_y = -(y_size / 2) - 10; /* Top alignment */
                    } else {
                        ref_pos_x = 10;
                        ref_pos_y = 15;
                        part_value_pos_x = 10;
                        part_value_pos_y = -y_size - 15;
                    }
                    if (opt_kicad) {
                        /* kicad: F0 "ref" X Y Size H V C CNN */
                        sprintf(stemp, "\nF0 \"%s\" %d %d 60 H V L BNN", ref, ref_pos_x, ref_pos_y); /* reference */
                        output_string(stemp);
                        /* kicad: F1 "name" X Y Size H V C CNN */
                        sprintf(stemp, "\nF1 \"%s\" %d %d 60 H V L TNN", partnames[0], part_value_pos_x,  part_value_pos_y); /* part name - include character height in y-offset calculation or use top alignment */
                        output_string(stemp);
                        /* kicad: F2 - footprint name, currently empty */
                        sprintf(stemp, "\nF2 \"\" 0 0 50 H I C CNN");
                        output_string(stemp);
                        /* kicad: F2 - datasheet filename, currently empty */
                        sprintf(stemp, "\nF3 \"\" 0 0 50 H I C CNN");
                        output_string(stemp);
                        components++;   /* statistics counter */
                        write_component(partnames[0], x_size, y_size); /* add component to component size file */
                        if (partnameidx > 1) {
                            sprintf(stemp, "\nALIAS");
                            output_string(stemp);
                            for (i=1; i<partnameidx; i++) {
                                sprintf(stemp, " %s", partnames[i]);
                                output_string(stemp);
                                write_component(partnames[i], x_size, y_size); /* add component to component size file */
                            }
                        }
                        sprintf(stemp, "\nDRAW");
                        output_string(stemp);
                    } else if (opt_geda) {
                        sprintf(stemp, "%s/%s-%1d.sym", lib_name, partnames[0], dmge);
                        partfile = fopen(stemp, "w");
                        if (partfile == NULL) {
                            printf("\n%s: %s part file create error", partnames, prog_name);;
                        } else {
                            /* Create relative symlinks to the file underway 
                             * and corresponding size entries
                             */
                            if (partnameidx > 1) {
                                /* symlink will point to the file in the same directory */
                                sprintf(stemp, "%s-%1d.sym", partnames[0], dmge);
                                for (int idx=1; idx<partnameidx; idx++) {
                                    sprintf(flink, "%s/%s-%1d.sym", lib_name, partnames[idx], dmge);
                                    unlink(flink); /* remove any old link, if present */
                                    symlink(stemp, flink); /* NOTE: ignore errors */
                                    write_component(partnames[idx], x_size, y_size); /* add component to component size file */
                                }
                            }
                        }
                        pin_sequence = 1; /* initialize pin sequence */
                        sprintf(stemp, "v 20031231 1");
                        write_part_data(stemp);
                        /* refdes: detached attribute color, visible, align bottom left */
                        sprintf(stemp, "\nT %d %d 8 6 1 1 0 0 1", ref_pos_x, ref_pos_y);
                        write_part_data(stemp);
                        sprintf(stemp, "\nrefdes=U?");
                        write_part_data(stemp);
                        /* partno: detached attribute color, visible, align top left */
                        sprintf(stemp, "\nT %d %d 8 6 1 1 0 2 1", part_value_pos_x,  part_value_pos_y);
                        write_part_data(stemp);
                        sprintf(stemp, "\npartno=%s", partnames[0]);
                        write_part_data(stemp);
                        /* device: detached attribute color, hidden */
                        sprintf(stemp, "\nT %d %d 8 3 0 0 0 0 1", 0,  0);
                        write_part_data(stemp);
                        sprintf(stemp, "\ndevice=%s", partnames[0]);
                        write_part_data(stemp);
                        if (parts_per_package) {
                            /* numslots: detached attribute color, hidden */
                            sprintf(stemp, "\nT %d %d 8 3 0 0 0 0 1", 200, 100);
                            write_part_data(stemp);
                            sprintf(stemp, "\nnumslots=%d", parts_per_package);
                            write_part_data(stemp);
                        }
                        components++;   /* statistics counter */
                        write_component(partnames[0], x_size, y_size); /* add component to component size file */
                        
                        if (parts_per_package > 1) {
                            /* 
                            * geda: Initialize the slot data array and indices 
                            */
                            for (i=0; i < MAXPARTSPERPACKAGE; i++) {
                                for (j=0; j < MAXPINSPERPART; j++) {
                                    slot_data[i][j] = 0;
                                }
                            }
                            gpins = 0;
                            gslots = 0;
                        }
                    }
                    enddraw_pending = 1;
                    
                    /* debug information is placed as comments (#) in the output file
                     * geda: cannot be the first line, where version is expected.
                     */
                    if (opt_verbose) {
                        sprintf(stemp, "\n# PARTNAME x_size: %d, y_size: %d, ppp: %d, inv: %d, gridarray: %d", \
                        x_size, y_size, parts_per_package, part_name_invisible, gridarray);
                        output_string(stemp);
                    }
                    processing_state = PINS; /* PARTNAME_COMPLETE -> PINS */
                }
        /* FALL THROUGH TO PINS */

        case PINS:
            /* First token is pin position and offset (Ln, Rn, Tn, Bn), n in 0.1 in,
             * followed by pin number columns, one for each part in the package [NOTE: may be empty], 
             * pin modifier(s) (SHORT, DOT, CLK,), 
             * pin type (IN, OUT, I/O, OC, PWR, PAS, hiZ), 
             * and pin name in single quotes; double quotes in pin name are removed for kicad
             * 
             * A single token denotes start of a bitmap, vector, or of another part (single quote).
             */
            if (opt_verbose) {sprintf(stemp, "\n# PINS state: %s", tokens[0]); output_string(stemp);}

            if (tokenidx == 0) {
                /* ignore empty lines, stay in the PINS state 
                 */
            } else if (tokenidx > 2) {
                if (opt_verbose) {sprintf(stemp, "\n# PINS state > 2: %s", tokens[0]); output_string(stemp);}
                /* Pin definition.
                 * Pin position (relative) evaluated first for calculating pin_x, pin_y
                 */
                pin_pos = atoi(&tokens[0][1]);
                
                /* Pin orientation
                 * same convention followed for kicad and geda:
                 * 0,0 is the top left corner of the component;
                 * x increases to the right, y increases upwards relative to anchor
                 * first pin position is the point away from the component
                 * kicad pin_orientation identifiers are also used for geda processing
                 */
                int pin_x, pin_y;
                pin_orientation[0] = 'L';
                pin_orientation[1] = '\0';
                switch (tokens[0][0]) {
                    case 'L':
                        pin_orientation[0] = 'R'; /* Left pin goes to the right */
                        break;
                    case 'R':
                        pin_orientation[0] = 'L'; /* Right pin goes to the left */
                        break;
                    case 'T':
                        pin_orientation[0] = 'D'; /* Top pin goes down */
                        break;
                    case 'B':
                        pin_orientation[0] = 'U'; /* Bottom pin goes up */
                        break;
                }

                /* Pin identification for GRIDARRAY (string) or pin numbers
                 */
                if (gridarray) {
                    tokenlen = strlen(tokens[1]) - 2;
                    strncpy(pin_gridarray, &tokens[1][1], tokenlen);
                    pin_gridarray[tokenlen] = '\0';
                } else {
                    i = 0;
                    do {
                        pin_number[i] = atoi(tokens[i+1]);
                         i++;
                    } while (i < parts_per_package);
                }
                
                /* pin attributes 
                 * NOTE: (i) is reused to index tokens for the next two code blocks
                 */
                int dot=0, clk=0, in=0, out=0, io=0, oc=0, pwr=0, pas=0, hiz=0, pin_short=0;
                char pin_type='P', pin_shape=0x20;
                i = parts_per_package + 1; /* next token to scan following pin numbers, with offset 0 */
                while (strstr(tokens[i], "\'") == NULL) { /* a quote indicates a pin name token */
                    if (strcmp(tokens[i], "DOT") == 0) {
                        dot=1;
                    } else if (strcmp(tokens[i], "CLK") == 0) {
                        clk=1;
                    } else if (strcmp(tokens[i], "IN") == 0) {
                        in=1; pin_type =  'I';
                    } else if (strcmp(tokens[i], "OUT") == 0) {
                        out=1; pin_type = 'O';
                    } else if (strcmp(tokens[i], "I/O") == 0) {
                        io=1; pin_type =  'B';
                    } else if (strcmp(tokens[i], "OC") == 0) {
                        oc=1; pin_type =  'C';
                    } else if (strcmp(tokens[i], "PWR") == 0) {
                        pwr=1; pin_type = 'W';      /* NOTE: Orcad does not distinguish between input or output power pin*/
                    } else if (strcmp(tokens[i], "PAS") == 0) {
                        pas=1; pin_type = 'P';
                    } else if (strcmp(tokens[i], "hiZ") == 0) {
                        hiz=1; pin_type = 'T';
                    } else if (strcmp(tokens[i], "SHORT") == 0) {
                        pin_short=1;
                    }
                    /* Pin shape: power is invisible, overrides all, clock overrides dot, dot (invert) */
                    if (pwr) {
                        pin_shape = 'N';
                    } else if (clk) {
                        pin_shape = 'C';
                    } else if (dot) {
                        pin_shape = 'I';
                    }
                    
                    i++;
                }
                /* pin name
                 * pin name should be more than two characters long.
                 * if empty, pin name is ''
                 * NOTE: token to scan (i) comes from previous code block 
                 */
                if (strlen(tokens[i]) > 2) {
//                strcpy(pin_name, tokens[i]);
                tokenlen = strlen(tokens[i]) - 2;
                strncpy(pin_name, &tokens[i][1], tokenlen);
                pin_name[tokenlen] = '\0';
                } else {
                    strcpy(pin_name, "~");  /* no pin name */
                }
                
                /* all pin parameters have been collected */
                if (opt_verbose) {
                    i = 0;
                    do {
                        sprintf(stemp, "\n# PINS def, part: %d -- %s %s %d %s dot %d clk %d in %d out %d io %d oc %d pwr %d pas %d hiz %d short %d", \
                            i+1, pin_name, pin_orientation, pin_pos, pin_gridarray, dot, clk, in, out, io, oc, pwr, pas, hiz, pin_short);
                        output_string(stemp);
                        i++;
                    } while (i<parts_per_package);
                }
                if (opt_kicad || opt_geda) {
                    i = 0;
                    do {
                        if (gridarray == 0) {
                            if (parts_per_package) {
                                sprintf(pin_gridarray, "%d", pin_number[i]);
                            } else {
                                /* parts_per_package = 0, no pin number */
                                sprintf(pin_gridarray, "~");
                            }
                        }
                        pin_length = ((pin_short==0)?300:100); /* mils */
                        
                        /* orcad: power pins do not show at all 
                         * geda:  power pins become nets (see below)
                         */
                        if (pwr) pin_length = 0;
                        
                        /* kicad: calculate the pin start position -- which lies outside of the component 
                         * kicad: 0,0 is top left corner of the component
                         * kicad: x increases to the right, y increases upwards 
                         * geda: need to calculate endpoint of pin, as well
                         * geda: adjust length to accommodate logic bubble (orcad: dot)
                         */
                        int pin_end_x, pin_end_y;
                        switch (pin_orientation[0]) {
                            case 'L':
                                pin_x = x_size + pin_length; /* Right pin goes to the left */
                                pin_y = -pin_pos * SCALE;
                                pin_end_x = dot?x_size + SCALE/2:x_size;
                                pin_end_y = pin_y;
                                break;
                            case 'R':
                                pin_x = -pin_length; /* Left pin goes to the right */
                                pin_y = -pin_pos * SCALE;
                                pin_end_x = dot?-SCALE/2:0;
                                pin_end_y = pin_y;
                                break;
                            case 'U':
                                pin_x = pin_pos * SCALE; /* Bottom pin goes up */
                                pin_y = -y_size - pin_length;
                                pin_end_x = pin_x;
                                pin_end_y = dot?-y_size-SCALE/2:-y_size;
                                break;
                            case 'D':
                                pin_x = pin_pos * SCALE; /* Top pin goes down */ 
                                pin_y = pin_length;
                                pin_end_x = pin_x;
                                pin_end_y = dot?SCALE/2:0;
                                break;
                        }
                        /* kicad: remove any double quotes from pin name
                         */
                        int idx=0;
                        while(pin_name[idx] != '\0') {
                            if(pin_name[idx] == '"') {
                                pin_name[idx] = '/';
                            }
                            idx++;
                        }
    
                        if (opt_kicad) {
                            /* handle backslashes (overbar)
                             * orcad: each character with overbar is followed by a backslash
                             * kicad: a tilde (~) marks beginning and optional end of the overbar 
                             */
                            unsigned char bs[MAXSTRING];
                            int bidx = 0;
                            int backslash = 0;
                            if (strstr(pin_name, "\\")) {
                                for (idx=0; idx<strlen(pin_name); idx++) {
                                    if (pin_name[idx] == '\\') {
                                        if (backslash == 2) {
                                            /* Ignore backslash if already under way */
                                            backslash = 1;
                                        } else {
                                            bs[bidx] = bs[bidx-1];
                                            bs[bidx-1] = '~';
                                            bs[++bidx] = '~'; /* store end tilde */
                                            backslash++;
                                        }
                                    } else {
                                            bs[bidx++] = pin_name[idx];
                                            if (backslash) backslash++;
                                    }
                                }
                                bs[bidx] = '\0';
                                strcpy(pin_name, bs); /* replace */
                            }
                        }

                        if (opt_kicad) {
                        /* X name pin X Y length orientation sizenum sizename part dmg type shape */
                        sprintf(stemp,"\nX %s %s %d %d %d %s 50 40 %d %d %c %c", \
                                pin_name, pin_gridarray, pin_x, pin_y, pin_length, pin_orientation, i+1, dmge, pin_type, pin_shape);
                        output_string(stemp);
                        } else if (opt_geda) {
                            /* 
                             * dot: draw a 50mil radius circle at package edge
                             * P x1 y1 x2 y2 color pintype whichend
                             * {
                             * T
                             * attributes (yellow: 5, except pinlabel: 3): 
                             * pinnumber (visible value), 
                             * pinseq (hidden, sequential, starting at 1), 
                             * pinlabel (visible value, green: 3), 
                             * pintype (hidden): in, out, io, oc, [oe], pas, [tp], tri (hiz), clk, pwr
                             * }
                             */
                            int align; /* used to align pin labels */
                            if (dot) {
                                int dot_x, dot_y;
                                switch (pin_orientation[0]) {
                                    case 'L':
                                        dot_x = pin_end_x - 25;
                                        dot_y = pin_end_y;
                                        break;
                                    case 'R':
                                        dot_x = pin_end_x + 25;
                                        dot_y = pin_end_y;
                                        break;
                                    case 'U':
                                        dot_x = pin_end_x;
                                        dot_y = pin_end_y + 25;
                                        break;
                                    case 'D':
                                        dot_x = pin_end_x;
                                        dot_y = pin_end_y - 25;
                                        break;
                                } 
                                sprintf(stemp,"\nV %d %d 25 6 0 0 0 -1 -1 0 -1 -1 -1 -1 -1", dot_x, dot_y);
                                write_part_data(stemp);
                            }

                            if (pwr && !opt_power) {
                                /* power pins are normally hidden and become nets with the same name
                                 * opt_power overrides this behavior and the pin is shown
                                 * (useful for power and discrete device libraries)
                                 */ 
                                sprintf(stemp, "\nT %d %d 5 3 0 0 0 0 1", pin_x, pin_y + 50);
                                write_part_data(stemp);
                                sprintf(stemp, "\nnet=%s:%s", pin_gridarray, pin_name);
                                write_part_data(stemp);
                            } else {
                                sprintf(stemp, "\nP %d %d %d %d 1 0 0", pin_x, pin_y, pin_end_x, pin_end_y);
                                write_part_data(stemp);
                                sprintf(stemp, "\n{");
                                write_part_data(stemp);
                                if (strcmp("~", pin_gridarray) == 0) {
                                    pin_gridarray[0]='\0';
                                }
                                /* Left pin data is right justified
                                 * Right pin data is left justified
                                 * Top and bottom pin data is centered
                                 */
                                if (pin_x <= 0) align = 6; else align = 0;
                                if (pin_x > 0 && pin_x < x_size && pin_y > 0) align = 5;
                                if (pin_x > 0 && pin_x < x_size && pin_y < -y_size) align = 3;
                                sprintf(stemp, "\nT %d %d 5 4 %d 1 0 %d 1", (align == 6)?pin_end_x - 80 + dot*55: pin_end_x + 80 - dot*50, pin_end_y + 10, part_name_invisible?0:1, align);
                                write_part_data(stemp);
                                sprintf(stemp, "\npinnumber=%s", pin_gridarray);
                                write_part_data(stemp);
                                sprintf(stemp, "\nT %d %d 5 3 0 0 0 0 1", pin_end_x, pin_end_y + 50);
                                write_part_data(stemp);
                                sprintf(stemp, "\npinseq=%d", pin_sequence++);
                                write_part_data(stemp);
                                if (strcmp("~", pin_name)  == 0) pin_name[0]='\0';
                                if (pin_x <= 0) align = 1; else align = 7;
                                sprintf(stemp, "\nT %d %d 3 4 1 1 0 %d 1", (align == 1)?pin_end_x + 50 + dot*55:pin_end_x - 50 - dot*50, pin_y, align);
                                write_part_data(stemp);
                                sprintf(stemp, "\npinlabel=%s", pin_name);
                                write_part_data(stemp);
                                sprintf(stemp, "\nT %d %d 5 3 0 0 0 0 1", pin_x, pin_y + 50);
                                write_part_data(stemp);
                                sprintf(stemp, "\npintype=");
                                write_part_data(stemp);
                                sprintf(stemp,"pas"); /* default type */
                                if (in) sprintf(stemp,"in");
                                else if (out) sprintf(stemp,"out");
                                else if (io) sprintf(stemp, "io");
                                else if (oc) sprintf(stemp, "oc");
                                else if (hiz) sprintf(stemp, "tri");
                                else if (clk) sprintf(stemp, "clk");
                                else if (pwr) sprintf(stemp, "pwr");
                                write_part_data(stemp);
                                sprintf(stemp,"\n}");
                                write_part_data(stemp);
                                
                                if (!pwr && parts_per_package > 1) {
                                    /* pin numbers for non-power pins are stored in corresponding slots
                                     * reset the slot index but don't touch the pin index
                                     */
                                    gslots = 0; 
                                    for (int k=0; k<parts_per_package; k++) {
                                        slot_data[gslots++][gpins] = pin_number[k];
                                    }
                                    gpins++; /* geda: point to next pin position */
                                }
                            }
                            i=parts_per_package; /* exit loop */
                        }
                        i++;
                    } while (i<parts_per_package);
                }
                
            } else if (strstr(tokens[0], "\'")) {
                if (opt_geda) write_slot_data(gpins); /* should not be needed here */
                if (opt_verbose) {sprintf(stemp, "\n# PINS -> PARTNAME: %s", tokens[0]); output_string(stemp);}

                /* partname detected -- combine all space-separated tokens */
                int spidx=1;
                while (tokenidx > spidx) {
                        strcat(tokens[0], "_");
                        strcat(tokens[0], tokens[spidx++]);
                }
                partnameidx = 0;
                gridarray = 0;
                tokenlen = strlen(tokens[0]) - 2;
                strncpy(partnames[partnameidx], &tokens[0][1], tokenlen);
                partnames[partnameidx][tokenlen] = '\0';
                subspaces(partnames[partnameidx]); /* kicad: spaces not allowed FIXME: for others*/
                NEWPART(1) 
                /* first part name already stored */
                processing_state = PARTNAME;
                
                if (enddraw_pending) {
                    if (opt_kicad) {
                        /* have not gone through the BITMAP or VECTOR state,
                        * must draw a rectangle.
                        * kicad: S X1 Y1 X2 Y2 part dmg pen fill Rectangle, from X1,Y1 to X2,Y2.
                        * kicad: 0,0 is top left corner of the component
                        * kicad: x increases to the right, y increases upwards 
                        * part = 0 (applies to all parts)
                        */
                        sprintf(stemp, "\nS 0 0 %d %d 0 %d 0 f", x_size, -y_size, dmge);
                        output_string(stemp);
                        sprintf(stemp, "\nENDDRAW");
                        output_string(stemp);
                        sprintf(stemp, "\nENDDEF");
                        output_string(stemp);
                    } else if (opt_geda) {
                        /* B x y width height color width capstyle dashstyle dashlength dashspace
                         *          filltype fillwidth angle1 pitch1 angle2 pitch2 */
                        sprintf(stemp, "\nB 0 0 %d %d 12 0 0 0 -1 -1 0 0 0 0 0 0", x_size, -y_size);
                        write_part_data(stemp);
                        if (partfile) {fclose(partfile); partfile = NULL;}
                    }
                    enddraw_pending = 0;
                } 
            } else if ( (tokenidx == 1) && (strcmp(tokens[0], "VECTOR") == 0)) {
                if (opt_verbose) {sprintf(stemp, "\n# PINS -> VECTOR (not expected): %s", tokens[0]); output_string(stemp);} 
                if (opt_geda) write_slot_data(gpins); /* should not be needed here */
                processing_state = VECTOR;
            } else if ( (tokenidx >= 2) && (strcmp(tokens[0], "BITMAP") == 0)) {
                if (opt_verbose) {sprintf(stemp, "\n# PINS -> BITMAP: %s %s", tokens[0], tokens[1]); output_string(stemp);}
                /* 
                 * Case of BITMAP "PARTNAME", which may include spaces
                 * All tokens have to be collected to generate a partname
                 * If the part is found in the existing_part table, 
                 * the input pointer is modified to point to the 
                 * corresponding VECTOR entry.
                 */
                unsigned char search_part[MAXSTRING];
                int expartidx;
                int spidx=2;
                while (tokenidx > spidx) {
                        strcat(tokens[1], "_");
                        strcat(tokens[1], tokens[spidx++]);
                }
                tokenlen = strlen(tokens[1]) - 2;
                strncpy(search_part, &tokens[1][1], tokenlen);
                search_part[tokenlen] = '\0';
                subspaces(search_part); /* kicad: spaces not allowed FIXME: for others*/
                /* Look up the name */
                expartidx = existing_vector(search_part);
                if (opt_verbose) {sprintf(stemp, "\n# PINS -> BITMAP: search_part %s expartidx %d", search_part, expartidx); output_string(stemp);}
                if (expartidx >=0) {
                    /* existing part - prepare for processing */
                    iptr_save = iptr;
                    iptr = vector_ptr[expartidx];
                    based_on_previous = 1; /* will be reset on VECTOR END directive */
                    if (opt_geda) write_slot_data(gpins);
                    processing_state = VECTOR;
                } else {
                    /* part not found - ignore directive - stay in this state */
                }

            } else if ( (tokenidx >= 2) && (strcmp(tokens[0], "CONVERT") == 0)) {
                if (opt_verbose) {sprintf(stemp, "\n# PINS -> CONVERT: %s %s", tokens[0], tokens[1]); output_string(stemp);}
                /* 
                 * Case of CONVERT "PARTNAME", which may include spaces
                 * All tokens have to be collected to generate a partname
                 * If the part is found in the existing_part table, 
                 * the input pointer is modified to point to the 
                 * corresponding VECTOR entry.
                 */
                unsigned char search_part[MAXSTRING];
                int expartidx;
                int spidx=2;
                while (tokenidx > spidx) {
                        strcat(tokens[1], "_");
                        strcat(tokens[1], tokens[spidx++]);
                }
                tokenlen = strlen(tokens[1]) - 2;
                strncpy(search_part, &tokens[1][1], tokenlen);
                search_part[tokenlen] = '\0';
                subspaces(search_part); /* kicad: spaces not allowed FIXME: for others*/
                /* Look up the name */
                expartidx = existing_vector(search_part);
                if (opt_verbose) {sprintf(stemp, "\n# PINS -> CONVERT: search_part %s expartidx %d", search_part, expartidx); output_string(stemp);}
                if (expartidx >=0) {
                    /* existing part - prepare for processing */
                    iptr_save = iptr;
                    iptr = convert_ptr[expartidx];
                    based_on_previous = 1; /* will be reset on VECTOR END directive */
                    if (opt_geda) write_slot_data(gpins);
                    processing_state = VECTOR;
                } else {
                    /* part not found - ignore directive - stay in this state */
                }

            } else if ( (tokenidx == 1) && (strstr(tokens[0], "#") || strstr(tokens[0], ".")) ) {
                if (opt_verbose) {sprintf(stemp, "\n# PINS -> BITMAP: %s", tokens[0]); output_string(stemp);}
                if (opt_geda) write_slot_data(gpins);
                processing_state = BITMAP;
            }
            break;
            
        case BITMAP:
            if (tokenidx == 0) {
                /* remain at the BITMAP state if empty line 
                 */
            } else if ( (tokenidx == 1) && (strcmp(tokens[0], "VECTOR") == 0)) {
                if (opt_verbose) {sprintf(stemp, "\n# BITMAP -> VECTOR+STORE: %s", tokens[0]); output_string(stemp);}
                /* Store VECTOR / CONVERT information location (iptr) */
                if (convert_state) {
                    convert_ptr[part_in_table] = iptr;
                } else {
                    strcpy(existing_part[part_in_table], partnames[0]);
                    vector_ptr[part_in_table] = iptr;
                    increment_part_in_table = 1;
                }
                /* part_in_table will be incremented on next VECTOR -> PARTNAME change */  
                processing_state = VECTOR;
            } else if ((tokenidx == 1) && (strcmp(tokens[0], "END") == 0)) { /* NOTE: this is an error */
                if (opt_verbose) {sprintf(stemp, "\n# BITMAP -> PARTNAME: %s", tokens[0]); output_string(stemp);}
                NEWPART(0)
                processing_state = PARTNAME;
            } else if (strstr(tokens[0],"#") || strstr(tokens[0],".") ) {
                if (opt_verbose) {
                    sprintf(stemp,"\n# BITMAP string: %s", tokens[0]);
                    output_string(stemp);
                }
            }
            break;
            
        case VECTOR:
            if (opt_verbose) {sprintf(stemp, "\n# VECTOR state: %s", tokens[0]); output_string(stemp);}
            if (tokenidx == 0) {
                /* Stay in the VECTOR state */
            } else if ( (tokenidx == 1) && (strcmp(tokens[0], "END") == 0) ) {
                if (based_on_previous) {
                    /* Restore input pointer */
                    based_on_previous = 0;
                    iptr = iptr_save;
                }
                if (opt_geda) {
                    /* Close primary component file "-1.sym"
                     */
                    if (partfile) {fclose(partfile); partfile = NULL;}
                }
                /* stay in the VECTOR state */
            } else if ( (tokenidx == 1) && (strcmp(tokens[0], "CONVERT") == 0) ) {
                if (opt_verbose) {sprintf(stemp, "\n# VECTOR -> PINS (convert): %s", tokens[0]); output_string(stemp);}
                dmge = 2; /* alternate (De Morgan) symbol */
                convert_state = 1;
                processing_state = PINS;
                if (opt_geda) {
                     /* Go to PARTNAME_COMPLETE state instead of PINS
                     * with the new dmge (2) and convert_state (1) values
                     */
                    if (partfile) {fclose(partfile); partfile = NULL;} /* NOTE: should already be closed */
                 
                    processing_state = PARTNAME_COMPLETE;
                }
            } else if (strstr(tokens[0], "\'")) {
              
                /* partname detected -- combine all space-separated tokens */
                int spidx=1;
                while (tokenidx > spidx) {
                        strcat(tokens[0], "_");
                        strcat(tokens[0], tokens[spidx++]);
                }
                partnameidx = 0;
                gridarray = 0;
                tokenlen = strlen(tokens[0]) - 2;
                strncpy(partnames[partnameidx], &tokens[0][1], tokenlen);
                partnames[partnameidx][tokenlen] = '\0';
                subspaces(partnames[partnameidx]); /* kicad: spaces not allowed, geda: spaces and hyphens not allowed */
                NEWPART(1)  
                /* first part name already stored */
                if (opt_kicad) {
                    sprintf(stemp, "\nENDDRAW");
                    output_string(stemp);
                    sprintf(stemp, "\nENDDEF");
                    output_string(stemp);
                    enddraw_pending = 0;
                }

                processing_state = PARTNAME;
            } else {
                if (strcmp(tokens[0], "LINE") == 0) {
                    if (opt_kicad) {
                       /* P count part dmg pen X Y ... fill 
                        * Polygon with count vertices, and an X,Y position for each vertex. 
                        * A filled polygon (fill = F, f) is implicitly closed, other polygons (fill = N) are open.
                        * part = 0 (applies to all parts)
                        * orcad: LINE X start, Y start, X end, Y end (relative to component 0, 0, Y increases downwards)
                        * orcad: count for LINE is always 2 and part = 0 (applies to all parts)
                        * kicad: x increases to the right, y increases upwards 
                        */
//                         if (0) {
//                             sprintf(stemp, "\n%s XS %f YS %f XE %f YE %f",\
//                             tokens[0], atof(tokens[1]), atof(tokens[2]), atof(tokens[3]), atof(tokens[4]));
//                             output_string(stemp);
//                         }
                        //FIXME: CHECK IF RESULTS ARE FLOATING
                        sprintf(stemp,"\nP 2 0 %d 0 %.0f %.0f %.0f %.0f N", \
                            dmge, rintf(atof(tokens[1])*SCALE),  -rintf(atof(tokens[2])*SCALE), \
                            rintf(atof(tokens[3])*SCALE),  -rintf(atof(tokens[4])*SCALE));
                        output_string(stemp);
                    } else if (opt_geda) {
                        /* L x1 y1 x2 y2 color width capstyle dashstyle dashlength dashspace */
                        sprintf(stemp,"\nL %.0f %.0f %.0f %.0f 3 0 0 0 -1 -1", \
                            rintf(atof(tokens[1])*SCALE),  -rintf(atof(tokens[2])*SCALE), \
                            rintf(atof(tokens[3])*SCALE),  -rintf(atof(tokens[4])*SCALE));
                        write_part_data(stemp);
                    }
                } else if (strcmp(tokens[0], "ARC") == 0) {
                    if (opt_kicad || opt_geda) {
                        double start_angle, end_angle, sweep_angle;
                        double center_x, center_y, start_x, start_y, end_x, end_y, radius;
                        /* kicad: A X Y radius start end part dmg pen fill Xstart Ystart Xend Yend
                        * part = 0 (applies to all parts)
                        * The start and end parameters are angles in 0.1 degrees. 
                        * The Xstart and Ystart parameters give the coordinate of the start point; 
                        * it can be calculated from the radius and the start angle. 
                        * Similarly, the Xend and Yend parameters give the coordinate of the end point. 
                        * The arc is drawn in counter-clockwise direction, 
                        * but the angles are swapped if their (normalized) difference exceeds 180 degrees.
                        * 
                        * orcad: ARC X_center, Y_center, X_start, Y_start, X_end, Y_end, radius 
                        *        (NOTE: X, Y start & end relative to arc center)
                        * kicad: Y is mirrored
                        * geda: integral degrees (x10)
                        */
//                         if (0) {
//                             sprintf(stemp, "\n%s XC %f YC %f XS %f YS %f XE %f YE %f radius %f",\
//                             tokens[0], atof(tokens[1]), atof(tokens[2]), atof(tokens[3]), atof(tokens[4]), atof(tokens[5]), atof(tokens[6]), atof(tokens[7]));
//                             output_string(stemp);
//                         }
                        center_x = rintf(atof(tokens[1]) * SCALE);
                        center_y = -(atof(tokens[2]) * SCALE);
                        start_x = rintf(atof(tokens[3]) * SCALE);
                        start_y = -rintf(atof(tokens[4]) * SCALE);
                        end_x = rintf(atof(tokens[5]) * SCALE);
                        end_y = -rintf(atof(tokens[6]) * SCALE);
                        radius = rintf(atof(tokens[7]) * SCALE);
                        start_angle = angle(center_x, center_y, center_x + start_x, center_y + start_y);
                        end_angle = angle(center_x, center_y, center_x + end_x, center_y + end_y);
//                         start_angle = angle(0., 0., 0., -200.);
//                         end_angle = angle(0., 0., 200., 0.);
//                         printf("\nstart: %f end: %f", start_angle, end_angle);
                        if (opt_kicad) {
                           sprintf(stemp, "\nA %.0f %.0f %.0f %.0f %.0f 0 %d 0 N %.0f %.0f %.0f %.0f",\
                           center_x, center_y, radius, start_angle, end_angle, dmge, center_x + start_x, center_y + start_y, center_x + end_x, center_y + end_y);
                           /* B count part dmg pen X Y ... fill Bezier curves with count vertices, and an X,Y position for each vertex. */
//                            sprintf(stemp, "\nB 2 0 %d 10 %.0f %.0f %.0f %.0f N",\
//                            center_x + start_x, center_y+ start_y, center_x + end_x, center_y + end_y);
                            output_string(stemp);
                        } else if (opt_geda) {
                            /* A x y radius startangle sweepangle color width capstyle dashstyle dashlength
                             * A 700 500 300 270 180 3 0 0 0 -1 -1 
                             */
                            start_angle = start_angle / 10.;
                            end_angle = end_angle / 10.;
                            sweep_angle = end_angle - start_angle;
                            if (sweep_angle > 180.) sweep_angle = sweep_angle - 360;
                            if (sweep_angle < -180.) sweep_angle = sweep_angle + 360;
                            sprintf(stemp, "\nA %.0f %.0f %.0f %.0f %.0f 3 0 0 0 -1 -1",\
                                center_x, center_y, radius, start_angle, sweep_angle);
                            write_part_data(stemp);
                        }
                    }
                } else if (strcmp(tokens[0], "CIRCLE") == 0) {
                    if (opt_kicad) {
                        /* kicad: C X Y radius part dmg pen fill 
                         * part = 0 (applies to all parts)
                         * orcad: CIRCLE X_center, Y_center, radius
                         */
                        sprintf(stemp,"\nC %.0f %.0f %.0f 0 %d 0 N", \
                            rintf(atof(tokens[1])*SCALE), -rintf(atof(tokens[2])*SCALE), rintf(atof(tokens[3])*SCALE), dmge);
                        output_string(stemp);
                    } else if (opt_geda) {
                        /* V x y radius color width capstyle dashstyle dashlength dashspace filltype fillwidth angle1 pitch1 angle2 pitch2
                         * V 1050 500 50 6 0 0 0 -1 -1 0 -1 -1 -1 -1 -1
                         */
                        sprintf(stemp,"\nV %.0f %.0f %.0f 3 0 0 0 -1 -1 0 -1 -1 -1 -1 -1", \
                            rintf(atof(tokens[1])*SCALE), -rintf(atof(tokens[2])*SCALE), rintf(atof(tokens[3])*SCALE));
                        write_part_data(stemp);
                    }
                } else if (strcmp(tokens[0], "TEXT") == 0) {
                    if (opt_kicad) {                        sprintf(stemp,"\nV %.0f %.0f %.0f 3 0 0 0 -1 -1 0 -1 -1 -1 -1 -1", \
                            rintf(atof(tokens[1])*SCALE), -rintf(atof(tokens[2])*SCALE), rintf(atof(tokens[3])*SCALE));
                        output_string(stemp);
                        /* kicad: T angle X Y size hidden part dmg text italic bold Halign Valign 
                         * orcad: TEXT X start, Y start, size, text in single quotes
                         * orcad: X start, Y start is lower left corner of text
                         * orcad: TEXT size from bitmaps appears to be 1 = 70mils, 2 = 140 mils
                         */
                        noticks(tokens[4]); /* remove single quotes from text */
                        sprintf(stemp,"\nT 0 %.0f %.0f %d 0 0 %d \"%s\" Normal 0 L B", \
                            rintf(atof(tokens[1])*SCALE), -rintf(atof(tokens[2])*SCALE), atoi(tokens[3])*70, dmge, tokens[4]);
                        output_string(stemp);
                    } else if (opt_geda) {
                            sprintf(stemp, "\nT %.0f %.0f 3 %d 1 1 0 0 1", \
                                    rintf(atof(tokens[1])*SCALE), -rintf(atof(tokens[2])*SCALE), atoi(tokens[3])*4);
                            write_part_data(stemp);
                            sprintf(stemp, "\n%s", tokens[4]);
                            write_part_data(stemp);
                    }
                }
            }
            break;
    }

}

void output_statistics(char *input_f)
{
    fprintf(stderr,"\n%s: Input file: %s", prog_name, input_f);
    if (part_in_table > 0) {
        fprintf(stderr,"\n%s: VECTOR/CONVERT Table - %d items\n", prog_name, --part_in_table);
    }
    for (int i=0; i < part_in_table; i++) {
        fprintf(stderr,"\n%s\tvector %lld\tconvert %lld", existing_part[i], vector_ptr[i], convert_ptr[i]);
    }
    
    fprintf(stderr,"\n\n%s: Total components: %d", prog_name, components);
    fprintf(stderr,"\n\n");
}
